"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.classes.loader import ClassLoader
from programy.dialog.sentence import Sentence
from programy.config.brain.securities import BrainSecuritiesConfiguration


class SecurityManager:

    def __init__(self, security_configuration):

        assert security_configuration is not None
        assert isinstance(security_configuration, BrainSecuritiesConfiguration)

        self._configuration = security_configuration
        self._authentication = None
        self._authorisation = None
        self._account_linker = None

    @property
    def authentication(self):
        return self._authentication

    @property
    def authorisation(self):
        return self._authorisation

    @property
    def account_linker(self):
        return self._account_linker

    def load_security_services(self, client):
        self.load_authentication_service(client)
        self.load_authorisation_service(client)
        self.load_account_linking_service(client)

    def _load_authentication_class(self, client):
        if self._configuration.authentication is not None:
            if self._configuration.authentication.classname is not None:
                classobject = ClassLoader.instantiate_class(self._configuration.authentication.classname)
                self._authentication = classobject(self._configuration.authentication)
                self._authentication.initialise(client)

    def load_authentication_service(self, client):
        try:
            self._load_authentication_class(client)

        except Exception as excep:
            YLogger.exception(self, "Failed to load security services", excep)

    def _load_authorisation_class(self, client):
        if self._configuration.authorisation is not None:
            if self._configuration.authorisation.classname is not None:
                classobject = ClassLoader.instantiate_class(self._configuration.authorisation.classname)
                self._authorisation = classobject(self._configuration.authorisation)
                self._authorisation.initialise(client)

    def load_authorisation_service(self, client):
        try:
            self._load_authorisation_class(client)

        except Exception as excep:
            YLogger.exception(self, "Failed to instatiate authorisation class", excep)

    def _load_account_linking_class(self, client):
        if self._configuration.account_linker is not None:
            if self._configuration.account_linker.classname is not None:
                classobject = ClassLoader.instantiate_class(self._configuration.account_linker.classname)
                self._account_linker = classobject(self._configuration.account_linker)
                self.account_linker.initialise(client)

    def load_account_linking_service(self, client):
        try:
            self._load_account_linking_class(client)

        except Exception as excep:
            YLogger.exception(self, "Failed to instatiate authorisation class", excep)

    def failed_authentication(self, client_context):
        YLogger.error(client_context, "[%s] failed authentication!")

        # If we have an SRAI defined, then use that
        if self.authentication.configuration.denied_srai is not None:
            match_context = client_context.brain.aiml_parser.\
                match_sentence(client_context,
                               Sentence(client_context,
                                        self.authentication.configuration.denied_srai),
                               topic_pattern="*",
                               that_pattern="*")

            # If the SRAI matched then return the result
            if match_context is not None:
                return client_context.brain.resolve_matched_template(client_context, match_context)

        # Otherswise return the static text, which is either
        #    User defined via config.yaml
        #    Or use the default value BrainSecurityConfiguration.DEFAULT_ACCESS_DENIED
        return self.authentication.configuration.denied_text

    def authenticate_user(self, client_context):
        if self.authentication is not None:
            if self.authentication.authenticate(client_context) is False:
                return self.failed_authentication(client_context)
        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.newsapi.newsapi import NewsAPI
from programy.extensions.base import Extension


class NewsAPIExtension(Extension):

    def get_news_api_api(self, context):
        return NewsAPI(context.client.license_keys)

    def get_news(self, context, source, max_num, sort, reverse):

        newsapi = self.get_news_api_api(context)

        headlines = newsapi.get_headlines(source, max_num, sort, reverse)
        if headlines is None:
            YLogger.error(context, "NewsAPIExtension no headlines found!")
            return ""

        results = newsapi.to_program_y_text(headlines)
        if results is None:
            YLogger.error(context, "NewsAPIExtension no results returned!")
            return ""

        return results

    def parse_data(self, context, data):
        source = None
        max_num = 10
        sort = False
        reverse = False

        splits = data.split()
        count = 0
        while count < len(splits):
            if splits[count] == "SOURCE":
                count += 1
                source = splits[count]
            elif splits[count] == "MAX":
                count += 1
                max_num = int(splits[count])
            elif splits[count] == "SORT":
                count += 1
                if splits[count].upper() == 'TRUE':
                    sort = True
                elif splits[count].upper() == 'FALSE':
                    sort = False
                else:
                    YLogger.error(context, "Invalid value for NewAPI Data parameter sort [%s]", splits[count])
                    sort = False
            elif splits[count] == "REVERSE":
                count += 1
                if splits[count].upper() == 'TRUE':
                    reverse = True
                elif splits[count].upper() == 'FALSE':
                    reverse = False
                else:
                    YLogger.error(context, "Invalid value for NewAPI Data parameter reverse [%s]", splits[count])
                    reverse = False
            else:
                YLogger.error(context, "Unknown News API Command [%s]", splits[count])

            count += 1

        return source, max_num, sort, reverse

    # execute() is the interface that is called from the <extension> tag in the AIML
    def execute(self, client_context, data):

        source, max_num, sort, reverse = self.parse_data(client_context, data)

        if source is None:
            YLogger.error(client_context, "NewsAPIExtension no source passed in as data parameter!")
            return ""

        return self.get_news(client_context, source, max_num, sort, reverse)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from programy.utils.logging.ylogger import YLogger

from programy.parser.template.nodes.base import TemplateNode
from programy.parser.template.nodes.get import TemplateGetNode
from programy.parser.template.nodes.bot import TemplateBotNode
from programy.parser.exceptions import ParserException
from programy.utils.text.text import TextUtils


class TemplateConditionVariable(TemplateNode):
    DEFAULT = 0
    GLOBAL = 1
    LOCAL = 2
    BOT = 3

    def __init__(self, name=None, value=None, var_type=GLOBAL, loop=False):
        TemplateNode.__init__(self)
        self._name = name
        self._value = value
        self._var_type = var_type
        self._loop = loop

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def var_type(self):
        return self._var_type

    @var_type.setter
    def var_type(self, var_type):
        self._var_type = var_type

    @property
    def loop(self):
        return self._loop

    @loop.setter
    def loop(self, loop):
        self._loop = loop

    def parse_expression(self, graph, expression):
        raise NotImplementedError("Never call this directly, call the subclass instead!")  # pragma: no cover


class TemplateConditionListItemNode(TemplateConditionVariable):

    def __init__(self, name=None, value=None, var_type=TemplateConditionVariable.GLOBAL, loop=False):
        TemplateConditionVariable.__init__(self, name, value, var_type, loop)

    def is_default(self):
        return bool(self.value is None)

    def to_string(self):
        valuestr = ""
        if self.value is not None:
            valuestr = self.value.to_string()

        if self.name is not None and self.value is not None:
            return "[CONDITIONLIST(%s=%s)]" % (self.name, valuestr)

        if self.name is not None:
            return "[CONDITIONLIST(%s)]" % self.name

        if self.value is not None:
            return "[CONDITIONLIST(%s)]" % valuestr

        return "[CONDITIONLIST]"

    def to_xml(self, client_context):

        xml = '<li'
        if self.name is not None:
            if self.var_type == TemplateConditionListItemNode.GLOBAL:
                xml += ' name="%s"' % self.name

            elif self.var_type == TemplateConditionListItemNode.LOCAL:
                xml += ' var="%s"' % self.name

            elif self.var_type == TemplateConditionListItemNode.BOT:
                xml += ' bot="%s"' % self.name

            elif self.var_type == TemplateConditionListItemNode.DEFAULT:
                xml += ' default="%s"' % self.name

            else:
                xml += ' unknown="%s"' % self.name

        xml += ">"

        if self.value is not None:
            xml += '<value>'
            xml += self.value.to_xml(client_context)
            xml += '</value>'

        xml += self.children_to_xml(client_context)

        if self.loop is True:
            xml += "<loop />"

        xml += '</li>'

        return xml

    def parse_expression(self, graph, expression):
        raise NotImplementedError("Never call this directly, call the subclass instead!")  # pragma: no cover


class TemplateConditionNode(TemplateConditionVariable):
    BLOCK = 1
    SINGLE = 2
    MULTIPLE = 3

    def __init__(self, name=None, value=None, var_type=TemplateConditionVariable.GLOBAL, loop=False,
                 condition_type=BLOCK):
        TemplateConditionVariable.__init__(self, name, value, var_type, loop)
        self._condition_type = condition_type

    def get_default(self):
        for child in self.children:
            if child.is_default() is True:
                return child
        return None

    #######################################################################################################
    # CONDITION_ITEM_COMPONENT ::== <name>TEMPLATE_EXPRESSION</name> | <value>TEMPLATE_EXPRESSION</value> | 
    #                                       <loop/> | TEMPLATE_EXPRESSION
    # CONDITION_ITEM_EXPRESSION ::== <li( CONDITION_ATTRIBUTES)*>(CONDITION_ITEM_COMPONENT)*</li>
    # CONDITION_ATTRIBUTES ::== (name="NAME") | (value="NORMALIZED_TEXT")
    # CONDITION_EXPRESSION ::== <condition( CONDITION_ATTRIBUTES)>(CONDITION_ITEM_EXPRESSION)*</condition>
    #

    def get_condition_name(self, condition):

        if 'name' in condition.attrib:
            return condition.attrib['name'], TemplateConditionVariable.GLOBAL

        elif 'var' in condition.attrib:
            return condition.attrib['var'], TemplateConditionVariable.LOCAL

        elif 'bot' in condition.attrib:
            return condition.attrib['bot'], TemplateConditionVariable.BOT

        else:
            names = condition.findall('name')
            if names:
                if len(names) > 1:
                    raise ParserException("Condition element has multiple name elements", xml_element=condition)
                name_text = self.get_text_from_element(condition.find('name'))
                return name_text, TemplateConditionVariable.GLOBAL

            variables = condition.findall('var')
            if variables:
                if len(variables) > 1:
                    raise ParserException("Condition element has multiple var elements", xml_element=condition)
                var_text = self.get_text_from_element(condition.find('var'))
                return var_text, TemplateConditionVariable.LOCAL

            bots = condition.findall('bot')
            if bots:
                if len(bots) > 1:
                    raise ParserException("Condition element has multiple bot elements", xml_element=condition)
                bot_text = self.get_text_from_element(condition.find('bot'))
                return bot_text, TemplateConditionVariable.BOT

        return None, TemplateConditionVariable.DEFAULT

    @staticmethod
    def get_condition_value(graph, condition):

        if 'value' in condition.attrib:
            value_node = graph.get_base_node()
            value_node.append(graph.get_word_node(condition.attrib['value']))
            return value_node

        else:
            values = condition.findall('value')
            if not values:
                return None

            elif len(values) > 1:
                raise ParserException("Element has multiple value elements", xml_element=condition)

            value_node = graph.get_base_node()
            value_node.parse_template_node(graph, values[0])
            return value_node

    # Type 1
    # <condition name="property" value="v">X</condition>,
    # <condition name="property"><value>v</value>X</condition>,
    # <condition value="v"><name>property</name>X</condition>, and
    # <condition><name>property</name><value>v</value>X</condition>
    #

    # Type 2
    # <condition name="property">
    # 	<li value="a">X</li>
    # 	<li value="b">Y</li>
    # 	<li>Z</li>				        <- Optional default value if no condition met
    # </ condition>
    # <condition name="property">
    # 	<li value="a">X</li>
    # 	<li value="b">Y</li>
    # 	<loop />				        <- Loop if no condition met
    # </ condition>
    # <condition name="property">
    # 	<li value="a">X</li>
    # 	<li value="b">Y <loop /></li>   <- Loop if condition set
    # </ condition>
    # or
    #
    # <condition>
    #   <name>property</name>
    # 	<li value="a">X</li>
    # 	<li value="b">Y</li>
    # 	<li>Z</li>				        <- Optional default value if no condition met
    # </ condition>
    #

    # Type 3
    #  <condition>
    # 	<li name='1' value="a">X</li>
    # 	<li value="b"><name>1</name>Y</li>
    # 	<li name="1"><value>b</value>Z</li>
    # 	<li><name>1</name><value>b</value>Z</li>
    # 	<li>Z<l/i>				        <- Optional default value if no condition met
    #  </condition>
    #

    def parse_expression(self, graph, expression):

        name, var_type = self.get_condition_name(expression)
        value = TemplateConditionNode.get_condition_value(graph, expression)

        if name is not None:
            self._name = name
            self._var_type = var_type
            if value is not None:
                self._condition_type = TemplateConditionNode.BLOCK
                self._value = value
                self.parse_type1_condition(graph, expression)

            else:
                self._condition_type = TemplateConditionNode.SINGLE
                self.parse_type2_condition(graph, expression)

        else:
            self._condition_type = TemplateConditionNode.MULTIPLE
            self.parse_type3_condition(graph, expression)

    def parse_type1_condition(self, graph, expression):
        self.parse_text(graph, self.get_text_from_element(expression))

        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name in ['name', 'var', 'bot', 'value']:
                pass

            elif tag_name == 'li':
                raise ParserException("li element not allowed as child of condition element",
                                      xml_element=expression)

            elif tag_name == 'loop':
                raise ParserException("This type of condition cannot have <loop /> element",
                                      xml_element=expression)

            else:
                graph.parse_tag_expression(child, self)

            tail_text = self.get_tail_from_element(child)
            self.parse_text(graph, tail_text)

    def parse_type2_condition(self, graph, expression):

        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name in ['name', 'var', 'bot']:
                pass

            elif tag_name == 'li':

                list_item = TemplateConditionListItemNode()

                list_item.value = TemplateConditionNode.get_condition_value(graph, child)
                list_item.var_type = self._var_type

                self.children.append(list_item)
                list_item.parse_text(graph, self.get_text_from_element(child))

                for sub_pattern in child:

                    if sub_pattern.tag in ['name', 'var', 'bot', 'value']:
                        pass

                    elif sub_pattern.tag == 'loop':
                        list_item.loop = True

                    else:
                        graph.parse_tag_expression(sub_pattern, list_item)

                    tail_text = self.get_tail_from_element(sub_pattern)
                    list_item.parse_text(graph, tail_text)

            else:
                raise ParserException("Invalid element <%s> in condition element" % tag_name,
                                      xml_element=expression)

    def parse_type3_condition(self, graph, expression):
        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)
            if tag_name == 'li':
                list_item = TemplateConditionListItemNode()

                name, var_type = self.get_condition_name(child)
                list_item.name = name
                list_item.var_type = var_type

                list_item.value = TemplateConditionNode.get_condition_value(graph, child)

                self.children.append(list_item)

                list_item.parse_text(graph, self.get_text_from_element(child))

                for sub_pattern in child:
                    if sub_pattern.tag in ['name', 'var', 'bot', 'value']:
                        pass

                    elif sub_pattern.tag == 'loop':
                        list_item.loop = True

                    else:
                        graph.parse_tag_expression(sub_pattern, list_item)

                    tail_text = self.get_tail_from_element(sub_pattern)
                    list_item.parse_text(graph, tail_text)

            else:
                raise ParserException("Invalid element <%s> in condition element" % tag_name, xml_element=expression)

    def to_string(self):
        text = "[CONDITION"
        if self.var_type == TemplateConditionListItemNode.GLOBAL:
            text += ' name="%s"' % self.name

        elif self.var_type == TemplateConditionListItemNode.LOCAL:
            text += ' var="%s"' % self.name

        elif self.var_type == TemplateConditionListItemNode.BOT:
            text += ' bot="%s"' % self.name

        elif self.var_type == TemplateConditionListItemNode.DEFAULT:
            text += ' default="%s"' % self.name

        else:
            text += ' unknown="%s"' % self.name

        if self.value is not None:
            text += " value=%s" % self.value

        text += "]"
        return text

    def to_xml(self, client_context):
        xml = "<condition"

        if self.name is not None:
            if self.var_type == TemplateConditionListItemNode.GLOBAL:
                xml += ' name="%s"' % self.name

            elif self.var_type == TemplateConditionListItemNode.LOCAL:
                xml += ' var="%s"' % self.name

            elif self.var_type == TemplateConditionListItemNode.BOT:
                xml += ' bot="%s"' % self.name

            elif self.var_type == TemplateConditionListItemNode.DEFAULT:
                xml += ' default="%s"' % self.name

            else:
                xml += ' unknown="%s"' % self.name

        xml += ">"

        if self.value is not None:
            xml += '<value>'
            xml += self.value.to_xml(client_context)
            xml += '</value>'

        xml += self.children_to_xml(client_context)
        xml += "</condition>"
        return xml

    def resolve(self, client_context):
        if self._condition_type == TemplateConditionNode.BLOCK:
            return self.resolve_type1_condition(client_context)

        elif self._condition_type == TemplateConditionNode.SINGLE:
            return self.resolve_type2_condition(client_context)

        else:
            return self.resolve_type3_condition(client_context)

    def get_condition_variable_value(self, client_context, var_type, name):
        if var_type == TemplateConditionVariable.GLOBAL:
            return TemplateGetNode.get_property_value(client_context, False, name)

        elif var_type == TemplateConditionVariable.LOCAL:
            return TemplateGetNode.get_property_value(client_context, True, name)

        else:
            return TemplateBotNode.get_bot_variable(client_context, name)

    def _resolve_type1_to_string(self, client_context):
        value = self.get_condition_variable_value(client_context, self.var_type, self.name)
        condition_value = self.value.resolve(client_context).upper()

        # Condition comparison is always case insensetive
        if value.upper() == condition_value:
            resolved = client_context.brain.tokenizer.words_to_texts([child.resolve(client_context)
                                                                      for child in self.children])
        else:
            resolved = ""

        YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)
        return resolved

    def resolve_type1_condition(self, client_context):
        try:
            return self._resolve_type1_to_string(client_context)

        except Exception as excep:
            YLogger.exception(client_context, "Failed to resolve type1 condition", excep)

        return ""

    def _resolve_type2_to_string(self, client_context):
        value = self.get_condition_variable_value(client_context, self.var_type, self.name)

        for condition in self.children:
            if condition.is_default() is False:
                condition_value = condition.value.resolve(client_context)

                # Condition comparison is always case insensetive
                if client_context.brain.tokenizer.compare(value.upper(), condition_value.upper()):
                    resolved = client_context.brain.tokenizer.words_to_texts([child_node.resolve(client_context)
                                                                              for child_node in condition.children])
                    YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)

                    if condition.loop is True:
                        self._pre_condition_loop_check(client_context)
                        resolved = resolved.strip() + " " + self.resolve(client_context)

                    return resolved

        default = self.get_default()
        if default is not None:
            resolved = client_context.brain.tokenizer.words_to_texts([child_node.resolve(client_context)
                                                                      for child_node in default.children])

            if default.loop is True:
                resolved = resolved.strip() + " " + self.resolve(client_context)
        else:
            resolved = ""

        YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)
        return resolved

    def resolve_type2_condition(self, client_context):
        try:
            return self._resolve_type2_to_string(client_context)

        except Exception as excep:
            YLogger.exception(client_context, "Failed to resolve type2 condition", excep)

        return ""

    def _resolve_type3_to_string(self, client_context):
        for condition in self.children:
            value = self.get_condition_variable_value(client_context, condition.var_type, condition.name)
            if condition.value is not None:
                condition_value = condition.value.resolve(client_context)

                # Condition comparison is always case insensetive
                if value.upper() == condition_value.upper():
                    resolved = client_context.brain.tokenizer.words_to_texts([child_node.resolve(client_context)
                                                                              for child_node in condition.children])
                    YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)

                    if condition.loop is True:
                        resolved = resolved.strip()
                        self._pre_condition_loop_check(client_context)
                        text = self.resolve(client_context)
                        resolved += " " + text.strip()

                    return resolved

        default = self.get_default()
        if default is not None:
            resolved = client_context.brain.tokenizer.words_to_texts([child_node.resolve(client_context)
                                                                      for child_node in default.children])

            if default.loop is True:
                resolved = resolved.strip()
                self._pre_default_loop_check(client_context)
                text = self.resolve(client_context)
                resolved += " " + text.strip()

        else:
            resolved = ""

        YLogger.debug(client_context, "[%s] resolved to [%s]", self.to_string(), resolved)
        return resolved

    def _pre_condition_loop_check(self, client_context):
        pass    #pragma: no cover

    def _pre_default_loop_check(self, client_context):
        pass    #pragma: no cover

    def resolve_type3_condition(self, client_context):
        try:
            return self._resolve_type3_to_string(client_context)

        except Exception as excep:
            YLogger.exception(client_context, "Failed to resolve type3 condition", excep)

        return ""



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.clients.config import ClientConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class SlackConfiguration(ClientConfigurationData):

    def __init__(self):
        ClientConfigurationData.__init__(self, "slack")
        self._polling_interval = 1
        self._username = "Chatilly"
        self._reply_to_direct = True
        self._reply_to_mention = True
        self._reply_to_all = False

    @property
    def polling_interval(self):
        return self._polling_interval

    @property
    def username(self):
        return self._username

    @property
    def reply_to_direct(self):
        return self._reply_to_direct

    @property
    def reply_to_mention(self):
        return self._reply_to_mention

    @property
    def reply_to_all(self):
        return self._reply_to_all

    def load_configuration_section(self, configuration_file, section, bot_root, subs: Substitutions = None):
        assert section is not None

        self._polling_interval = configuration_file.get_int_option(section, "polling_interval", missing_value=1,
                                                                   subs=subs)
        self._username = configuration_file.get_option(section, "username", missing_value="Chatilly",
                                                                   subs=subs)

        self._reply_to_direct = configuration_file.get_bool_option(section, "reply_to_direct", missing_value=True,
                                                                   subs=subs)
        self._reply_to_mention = configuration_file.get_bool_option(section, "reply_to_mention", missing_value=True,
                                                                   subs=subs)
        self._reply_to_all = configuration_file.get_bool_option(section, "reply_to_all", missing_value=False,
                                                                   subs=subs)

        super(SlackConfiguration, self).load_configuration_section(configuration_file, section, bot_root, subs=subs)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['polling_interval'] = 1
            data['username'] = "Chatilly"
            data['reply_to_direct'] = True
            data['reply_to_mention'] = True
            data['reply_to_all'] = True
        else:
            data['polling_interval'] = self._polling_interval
            data['username'] = self._username
            data['reply_to_direct'] = self._reply_to_direct
            data['reply_to_mention'] = self._reply_to_mention
            data['reply_to_all'] = self._reply_to_all

        super(SlackConfiguration, self).to_yaml(data, defaults)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import datetime


class DateFormatter:

    @staticmethod
    def year_month_day(year, month, day):
        return datetime.datetime.strptime("%d-%d-%d" % (year, month, day), "%Y-%m-%d")

    @staticmethod
    def year_month_day_now():
        now = datetime.datetime.now()
        return datetime.datetime.strptime(now.strftime("%Y-%m-%d"), "%Y-%m-%d")

    def __init__(self, weeks=0, days=0, hours=0, minutes=0, seconds=0):
        self._time_now = datetime.datetime.now()
        if weeks > 0:
            new_now = self._time_now + datetime.timedelta(days=weeks * 7)
            self._time_now = new_now
        if days > 0:
            new_now = self._time_now + datetime.timedelta(days=days)
            self._time_now = new_now
        if hours > 0:
            new_now = self._time_now + datetime.timedelta(seconds=hours * 60 * 60)
            self._time_now = new_now
        if minutes > 0:
            new_now = self._time_now + datetime.timedelta(seconds=minutes * 60)
            self._time_now = new_now
        if seconds > 0:
            new_now = self._time_now + datetime.timedelta(seconds=seconds)
            self._time_now = new_now

    @property
    def time_now(self):
        return self._time_now

    def abbreviated_weekday(self):
        return self._time_now.strftime("%a")

    def full_weekday(self):
        return self._time_now.strftime("%A")

    def abbreviated_month(self):
        return self._time_now.strftime("%b")

    def full_month(self):
        return self._time_now.strftime("%B")

    def locate_appropriate_date_time(self):
        return self._time_now.strftime("%c")

    def decimal_day_of_month(self):
        return self._time_now.strftime("%d")

    def hour_24_hour_clock(self):
        return self._time_now.strftime("%H")

    def hour_12_hour_clock(self):
        return self._time_now.strftime("%I")

    def decimal_day_of_year(self):
        return self._time_now.strftime("%j")

    def decimal_month(self):
        return self._time_now.strftime("%m")

    def decimal_minute(self):
        return self._time_now.strftime("%M")

    def am_or_pm(self):
        return self._time_now.strftime("%p")

    def decimal_second(self):
        return self._time_now.strftime("%S")

    def decimal_week_number_sunday_as_first(self):
        return self._time_now.strftime("%U")

    def decimal_week_number_monday_as_first(self):
        return self._time_now.strftime("%W")

    def decimal_weekday(self):
        return self._time_now.strftime("%w")

    def date_representation(self):
        return self._time_now.strftime("%x")

    def time_representation(self):
        return self._time_now.strftime("%X")

    def year_2_digit(self):
        return self._time_now.strftime("%y")

    def year_4_digit(self):
        return self._time_now.strftime("%Y")

    def timezone_name(self):
        return self._time_now.strftime("%Z")



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.security.authorise.usergroups import User
from programy.security.authorise.usergroups import Group
from programy.storage.entities.store import Store


class UserGroupsStore(Store):

    def __init__(self):
        Store.__init__(self)

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):
        raise NotImplementedError()  # pragma: no cover

    def load_usergroups(self, usersgroupsauthorisor):
        raise NotImplementedError()  # pragma: no cover

    def load_users_and_groups_from_yaml(self, yaml_data, usergroups):
        self._load_users(yaml_data, usergroups)
        self._load_groups(yaml_data, usergroups)
        self._combine_users_and_groups(usergroups)

    def _load_users(self, yaml_data, usergroups):
        if 'users' in yaml_data:
            for user_name in yaml_data['users'].keys():
                self._load_users_user(yaml_data, user_name, usergroups)

    def _load_users_user(self, yaml_data, user_name, usergroups):
        user = User(user_name)

        yaml_obj = yaml_data['users'][user_name]
        if 'roles' in yaml_obj:
            self._load_users_user_roles(yaml_obj, user, user_name)

        if 'groups' in yaml_obj:
            self._load_users_user_groups(yaml_obj, user, user_name)

        usergroups.users[user.userid] = user

    def _load_users_user_roles(self, yaml_obj, user, user_name):
        roles_list = yaml_obj['roles']
        splits = roles_list.split(",")
        for role_name in splits:
            role_name = role_name.strip()
            if role_name not in user.roles:
                user.roles.append(role_name)
            else:
                YLogger.debug(self, "Role [%s] already exists in user [%s]", role_name, user_name)

    def _load_users_user_groups(self, yaml_obj, user, user_name):
        groups_list = yaml_obj['groups']
        splits = groups_list.split(",")
        for group_name in splits:
            group_name = group_name.strip()
            if group_name not in user.groups:
                user.groups.append(group_name)
            else:
                YLogger.debug(self, "Group [%s] already exists in user [%s]", group_name, user_name)

    def _load_groups(self, yaml_data, usergroups):
        if 'groups' in yaml_data:
            for group_name in yaml_data['groups'].keys():
                self._load_groups_group(yaml_data, group_name, usergroups)

    def _load_groups_group(self, yaml_data, group_name, usergroups):
        group = Group(group_name)

        yaml_obj = yaml_data['groups'][group_name]
        if 'roles' in yaml_obj:
            self._load_groups_group_roles(yaml_obj, group, group_name)

        if 'groups' in yaml_obj:
            self._load_groups_group_groups(yaml_obj, group, group_name)

        if 'users' in yaml_obj:
            self._load_groups_group_users(yaml_obj, group, group_name)

        usergroups.groups[group.groupid] = group

    def _load_groups_group_roles(self, yaml_obj, group, group_name):
        roles_list = yaml_obj['roles']
        splits = roles_list.split(",")
        for role_name in splits:
            role_name = role_name.strip()
            if role_name not in group.roles:
                group.roles.append(role_name)
            else:
                YLogger.debug(self, "Role [%s] already exists in group [%s]", role_name, group_name)

    def _load_groups_group_groups(self, yaml_obj, group, group_name):
        groups_list = yaml_obj['groups']
        splits = groups_list.split(",")
        for element in splits:
            inner_group_name = element.strip()
            if inner_group_name not in group.groups:
                group.groups.append(inner_group_name)
            else:
                YLogger.debug(self, "Group [%s] already exists in group [%s]", inner_group_name, group_name)

    def _load_groups_group_users(self, yaml_obj, group, group_name):
        users_list = yaml_obj['users']
        splits = users_list.split(",")
        for user_name in splits:
            user_name = user_name.strip()
            if user_name not in group.users:
                group.users.append(user_name)
            else:
                YLogger.debug(self, "User [%s] already exists in group [%s]", user_name, group_name)

    def _combine_users_and_groups(self, usergroups):

        for user_id in usergroups.users.keys():
            user = usergroups.users[user_id]
            self._combine_user_groups(usergroups, user, user_id)

        for group_id in usergroups.groups.keys():
            group = usergroups.groups[group_id]
            self._combine_group_groups(usergroups, group, group_id)
            self._combine_group_users(usergroups, group, group_id)

    def _combine_user_groups(self, usergroups, user, user_id):
        new_groups = []
        for group_id in user.groups:
            if group_id in usergroups.groups:
                group = usergroups.groups[group_id]
                new_groups.append(group)
            else:
                YLogger.error(self, "Unknown group id [%s] in user [%s]", group_id, user_id)
        user.add_groups(new_groups[:])

    def _combine_group_groups(self, usergroups, group, group_id):
        new_groups = []
        for sub_group_id in group.groups:
            if sub_group_id in usergroups.groups:
                new_group = usergroups.groups[sub_group_id]
                new_groups.append(new_group)
            else:
                YLogger.error(self, "Unknown group id [%s] in group [%s]", sub_group_id, group_id)
        group.add_groups(new_groups[:])

    def _combine_group_users(self, usergroups, group, group_id):
        new_users = []
        for sub_user_id in group.users:
            if sub_user_id in usergroups.users:
                new_user = usergroups.users[sub_user_id]
                new_users.append(new_user)
            else:
                YLogger.error(self, "Unknown user id [%s] in group [%s]", sub_user_id, group_id)
        group.add_users(new_users[:])




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import os.path
import re
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.sql.store.sqlstore import SQLStore
from programy.storage.entities.property import PropertyStore
from programy.storage.stores.sql.dao.property import Property
from programy.storage.stores.sql.dao.property import DefaultVariable
from programy.storage.stores.sql.dao.property import Regex
from programy.mappings.base import DoubleStringPatternSplitCollection
from programy.storage.entities.store import Store


class SQLBasePropertyStore(SQLStore, PropertyStore):
    SPLIT_CHAR = ':'
    COMMENT = '#'

    def __init__(self, storage_engine):
        SQLStore.__init__(self, storage_engine)
        PropertyStore.__init__(self)

    def empty(self):
        self._get_all().delete()

    def empty_properties(self):
        self._get_all().delete()

    def _get_entity(self, name, value):
        raise NotImplementedError()  # pragma: no cover

    def add_property(self, name, value):
        prop = self._get_entity(name=name, value=value)
        self._storage_engine.session.add(prop)
        return True

    def add_properties(self, properties):
        prop_list = []
        for name, value in properties.items():
            prop_list.append(self._get_entity(name=name, value=value))
        self._storage_engine.session.add_all(prop_list)
        return True

    def get_properties(self):
        db_properties = self._get_all()
        properties = {}
        for prop in db_properties:
            properties[prop.name] = prop.value
        return properties

    def load(self, collector, name=None):
        del name
        self.load_all(collector)

    def load_all(self, collector):
        collector.empty()
        db_propertys = self._get_all()
        for db_property in db_propertys:
            self.add_to_collection(collector, db_property.name, db_property.value)

    def add_to_collection(self, collection, name, value):
        collection.add_property(name, value)

    def split_into_fields(self, line):
        return DoubleStringPatternSplitCollection.\
            split_line_by_pattern(line, DoubleStringPatternSplitCollection.RE_OF_SPLIT_PATTERN)

    def _read_lines_from_file(self, filename, verbose):
        count = 0
        success= 0
        with open(filename, "r") as vars_file:
            for line in vars_file:
                line = line.strip()
                if line.startswith(SQLBasePropertyStore.COMMENT) is False:
                    splits = line.split(SQLBasePropertyStore.SPLIT_CHAR)
                    if len(splits) > 1:
                        key = splits[0].strip()
                        val = ":".join(splits[1:]).strip()
                        self.add_property(key, val)
                        success += 1
                count += 1

        return count, success

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):
        try:
            count, success = self._read_lines_from_file(filename, verbose)

            self.commit(commit)

            return count, success

        except Exception as error:
            YLogger.exception(self, "Failed to load properties from [%s]", error, filename)

        return 0, 0


class SQLPropertyStore(SQLBasePropertyStore, PropertyStore):

    def __init__(self, storage_engine):
        SQLBasePropertyStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(Property)

    def _get_entity(self, name, value):
        return Property(name=name, value=value)


class SQLDefaultVariableStore(SQLBasePropertyStore, PropertyStore):

    def __init__(self, storage_engine):
        SQLBasePropertyStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(DefaultVariable)

    def _get_entity(self, name, value):
        return DefaultVariable(name=name, value=value)

    def add_default(self, name, value):
        return self.add_property(name, value)

    def add_defaults(self, defaults):
        self.add_properties(defaults)

    def get_default_values(self):
        return self.get_properties()


class SQLRegexStore(SQLBasePropertyStore, PropertyStore):

    def __init__(self, storage_engine):
        SQLBasePropertyStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(Regex)

    def _get_entity(self, name, value):
        return Regex(name=name, value=value)

    def add_regex(self, name, value):
        return self.add_property(name, value)

    def add_regexes(self, defaults):
        self.add_properties(defaults)

    def get_regexes(self):
        return self.get_properties()

    def add_to_collection(self, collection, name, value):
        try:
            collection.add_regex(name, re.compile(value, re.IGNORECASE))

        except Exception as excep:
            YLogger.exception_nostack(self, "Error adding regex to collection: [%s]" % value, excep)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.config.base import BaseConfigurationData
from programy.storage.stores.sql.engine import SQLStorageEngine
from programy.utils.substitutions.substitues import Substitutions


class SQLStorageConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="config")

        self._url = 'sqlite:///:memory:'
        self._echo = False
        self._encoding = 'utf-8'
        self._create_db = True
        self._drop_all_first = True

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._url = url

    @property
    def echo(self):
        return self._echo

    @property
    def encoding(self):
        return self._encoding

    @property
    def create_db(self):
        return self._create_db

    @create_db.setter
    def create_db(self, create):
        self._create_db = create

    @property
    def drop_all_first(self):
        return self._drop_all_first

    @drop_all_first.setter
    def drop_all_first(self, drop_all):
        self._drop_all_first = drop_all

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        storage = configuration_file.get_section(self._section_name, configuration)
        if storage is not None:
            self._url = configuration_file.get_option(storage, "url", subs=subs)
            if self._url.endswith("memory"):
                self._url += ":"
            self._echo = configuration_file.get_option(storage, "echo", subs=subs)
            self._encoding = configuration_file.get_option(storage, "encoding", subs=subs)
            self._create_db = configuration_file.get_option(storage, "create_db", subs=subs)
            self._drop_all_first = configuration_file.get_option(storage, "drop_all_first", subs=subs)

        else:
            YLogger.error(None, "'config' section missing from storage config")

    def create_sqlstorage_config(self):
        return {'url': self._url,
                'echo': self._echo,
                'encoding': self._encoding,
                'create_db': self._create_db,
                'drop_all_first': self._drop_all_first}

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['url'] = 'sqlite:///:memory:'
            data['echo'] = False
            data['encoding'] = 'utf-8'
            data['create_db'] = True
            data['drop_all_first'] = True
        else:
            data['url'] = self._url
            data['echo'] = self._echo
            data['encoding'] = self._encoding
            data['create_db'] = self._create_db
            data['drop_all_first'] = self._drop_all_first

    def create_engine(self):
        engine = SQLStorageEngine(self)
        engine.initialise()
        return engine



import xml.etree.ElementTree as ET
import sys

def load_replacements(replace_file):
    texts = {}
    bots = {}
    sets = {}
    with open(replace_file, "r") as file:
        for line in file:
            line = line.strip()
            name_value = line.split(":")
            type_name = name_value[0].split(",")
            if len(type_name) == 2:
                if type_name[0] == "SET":
                    sets[type_name[1]] = name_value[1].upper()
                elif type_name[0] == "BOT":
                    bots[type_name[1]] = name_value[1].upper()
                else:
                    print("Unknown mapping type", type_name[0].upper())
            else:
                texts[type_name[0]] = name_value[1]

    return texts, sets, bots

def replace_wildcard(text, texts, wildcard):
    if wildcard in text and wildcard in texts:
        return text.replace(wildcard, texts[wildcard])
    return text

def replace_wildcards(text, texts):
    text = replace_wildcard(text, texts, "*")
    text = replace_wildcard(text, texts, "#")
    text = replace_wildcard(text, texts, "^")
    text = replace_wildcard(text, texts, "_")
    return text

if __name__ == '__main__':

    aiml_file = sys.argv[1]
    test_file = sys.argv[2]
    ljust = int(sys.argv[3])
    replace_file = sys.argv[4]

    default = None
    if len(sys.argv) > 5:
        default = sys.argv[5]

    print("aiml_file:", aiml_file)
    print("test_file:", test_file)
    print("replace_file:", replace_file)
    print("Default:", default)

    texts, sets, bots = load_replacements(replace_file)

    with open(test_file, "w+") as output_file:
        if default is not None:
            output_file.write('$DEFAULT, "%s"\n\n'%default)
        try:
            tree = ET.parse(aiml_file)
            aiml = tree.getroot()
            categories = aiml.findall('category')
            for category in categories:
                pattern_text = ""

                pattern = category.find("pattern")
                for elt in pattern.iter():

                    if elt.tag == "pattern":
                        text = elt.text.strip().upper()
                        pattern_text += replace_wildcards(text, texts)

                    elif elt.tag == "set":
                        if 'name' in elt.attrib:
                            name = elt.attrib['name']
                        else:
                            name = elt.text.strip()

                        if name in sets:
                            pattern_text += sets[name]
                        else:
                            pattern_text += "SET[%s]"%name

                    elif elt.tag == "bot":
                        if 'name' in elt.attrib:
                            name = elt.attrib['name']
                        else:
                            name = elt.text.strip()

                        if name in bots:
                            pattern_text += bots[name]
                        else:
                            pattern_text += "BOT[%s]" % name

                    pattern_text += " "

                    if elt.tail is not None and elt.tail.strip() != "":
                        text = elt.tail.strip().upper()
                        pattern_text += replace_wildcards(text, texts)
                        pattern_text += " "

                question = '"%s",'%pattern_text.strip()
                question = question.ljust(ljust)


                if default is not None:
                    test_line = '%s $DEFAULT'%(question)

                else:
                    template = category.find('template')
                    test_line = '%s "answer"'%(question)

                output_file.write(test_line)
                output_file.write("\n")

            topics = aiml.findall('topic')
            if len(topics) > 0:
                print("I dont handle topics yet!")

        except Exception as e:
            print(e)

        exit(0)






import re
import pickle
import sys
from collections import Counter

if __name__ == '__main__':

    def all_words(text):
        return re.findall(r'\w+', text.upper())

    def create_corpus(text_filename, corpus_filename):

        words = Counter(all_words(open(text_filename).read()))
        sum_of_words = sum(words.values())

        corpus = (words, sum_of_words)
        with open(corpus_filename, "wb+") as f:
            pickle.dump(corpus, f)

    input_file = sys.argv[1]
    output_file = sys.argv[2]

    create_corpus(input_file, output_file)


import sys
import os.path

if __name__ == '__main__':

    rdf_dir = sys.argv[1]
    csv_file = sys.argv[2]

    print("rdf_dir:", rdf_dir)
    print("csv_file:", csv_file)

    entities = []

    files = 0
    for dirpath, dirnames, filenames in os.walk(rdf_dir):
        for filename in filenames:
            files += 1
            rdf_file = os.path.join(dirpath, filename)
            file = open(rdf_file, "r")
            for line in file:
                if line and len(line.strip()) > 0:
                    entity = []
                    rdf_splits = line.split(":")
                    entity.append(rdf_file)
                    entity += rdf_splits
                    entities.append(entity)
            file.close ()

    entities.sort(key=lambda x: x[1])

    with open(csv_file, "w+") as output_file:
        for entity in entities:
            new_line = ", ".join(entity)
            output_file.write(new_line)

    print("Files: %d"%files)
    print("RDFs:  %d"%len(entities))


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.pattern.nodes.base import PatternNode
from programy.parser.pattern.equalsmatch import EqualsMatch
from programy.parser.exceptions import ParserException


class PatternSetNode(PatternNode):

    def __init__(self, attribs, text, userid='*'):
        self._set_name = None
        self._additional = {}

        PatternNode.__init__(self, userid)
        for attrib in attribs:
            if attrib == 'name':
                self._set_name = attribs['name'].upper()
            else:
                self._additional[attrib] = attribs[attrib].upper()

        if text:
            self._set_name = text.upper()

        if self._set_name is None:
            raise ParserException("Invalid set node, no name specified as attribute or text")

    @property
    def set_name(self):
        return self._set_name

    @property
    def additional(self):
        return self._additional

    def is_set(self):
        return True

    def to_xml(self, client_context, include_user=False):
        string = ""
        if include_user is True:
            string += '<set userid="%s" name="%s"' % (self.userid, self.set_name)
        else:
            string += '<set name="%s"' % self.set_name

        for name, value in self._additional.items():
            string += ' %s="%s"' % (name, value)

        string += ">\n"
        string += super(PatternSetNode, self).to_xml(client_context)
        string += "</set>"

        return string

    def to_string(self, verbose=True):
        if verbose is True:
            string = "SET [%s] [%s] name=[%s]" % (self.userid, self._child_count(verbose), self.set_name)
        else:
            string = "SET name=[%s]" % (self.set_name)

        for name, value in self._additional.items():
            string += ' %s=[%s]' % (name, value)

        return string

    def set_is_numeric(self):
        return bool(self.set_name.upper() == 'NUMBER')

    def set_is_known(self, client_context):
        return bool(client_context.brain.sets.contains(self.set_name))

    def words_in_set(self, client_context, words, word_no):

        word = words.word(word_no).upper()
        set_words = client_context.brain.sets.set(self.set_name)
        if not set_words:
            YLogger.error(self, "No set with name [%s]", self.set_name)

        elif word in set_words:
            phrases = set_words[word]
            phrases = sorted(phrases, key=len, reverse=True)
            for phrase in phrases:
                phrase_word_no = 0
                words_word_no = word_no
                while phrase_word_no < len(phrase) and words_word_no < words.num_words():
                    phrase_word = phrase[phrase_word_no].upper()
                    word = words.word(words_word_no).upper()
                    if phrase_word == word:
                        if phrase_word_no + 1 == len(phrase):
                            return EqualsMatch(True, words_word_no, " ".join(phrase))
                    phrase_word_no += 1
                    words_word_no += 1

        return EqualsMatch(False, word_no)

    def equivalent(self, other):
        if other.is_set():
            if self.userid == other.userid:
                if self.set_name == other.set_name:
                    return True
        return False

    def equals(self, client_context, words, word_no):
        word = words.word(word_no)

        if self.userid != '*':
            if self.userid != client_context.userid:
                return EqualsMatch(False, word_no)

        if client_context.brain.dynamics.is_dynamic_set(self._set_name) is True:
            result = client_context.brain.dynamics.dynamic_set(client_context, self._set_name, word, self._additional)
            return EqualsMatch(result, word_no, word)

        else:
            if self.set_is_known(client_context):
                match = self.words_in_set(client_context, words, word_no)
                if match.matched is True:
                    YLogger.debug(client_context, "Found word [%s] in set [%s]", word, self.set_name)
                    return match

                else:
                    YLogger.debug(client_context, "No word [%s] found in set [%s]", word, self.set_name)
                    return EqualsMatch(False, word_no)

            else:
                YLogger.error(client_context, "No set named [%s] in sets collection", self.set_name)
                return EqualsMatch(False, word_no)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
# Ignore pylint warning, this import from Programy must be before ElementTree
# Which ensures that the class LineNumberingParser is injected into the code
from programy.utils.parsing.linenumxml import LineNumberingParser
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-order
from programy.utils.logging.ylogger import YLogger
from programy.config.file.file import BaseConfigurationFile
from programy.config.programy import ProgramyConfiguration
from programy.utils.substitutions.substitues import Substitutions


class XMLConfigurationFile(BaseConfigurationFile):

    def __init__(self):
        BaseConfigurationFile.__init__(self)
        self.xml_data = None

    def load_from_text(self, text, client_configuration, bot_root, subs: Substitutions = None):
        tree = ET.fromstring(text)
        self.xml_data = tree
        configuration = ProgramyConfiguration(client_configuration)
        configuration.load_config_data(self, bot_root, subs)
        return configuration

    def load_from_file(self, filename, client_configuration, bot_root, subs: Substitutions = None):
        configuration = ProgramyConfiguration(client_configuration)

        try:
            with open(filename, 'r+', encoding="utf-8") as xml_data_file:
                tree = ET.parse(xml_data_file, parser=LineNumberingParser())
                self.xml_data = tree.getroot()
                configuration.load_config_data(self, bot_root, subs)

        except Exception as excep:
            YLogger.exception(self, "Failed to open xml config file [%s]", excep, filename)

        return configuration

    def get_section(self, section_name, parent_section=None):
        if parent_section is None:
            return self.xml_data.find(section_name)
        return parent_section.find(section_name)

    def get_keys(self, section):
        keys = []
        for child in section:
            keys.append(child.tag)
        return keys

    def get_child_section_keys(self, child_section_name, parent_section):
        found = parent_section.find(child_section_name)
        if found is not None:
            keys = []
            for child in found:
                keys.append(child.tag)
            return keys
        return None

    def get_option(self, section, option_name, missing_value=None, subs=None):
        child = section.find(option_name)
        if child is not None:
            has_children = False
            for _ in child:
                has_children = True
                break

            if has_children is False:
                value = self._replace_subs(subs, child.text)
                return self._infer_type_from_string(value)

            return child

        if missing_value is not None:
            YLogger.warning(self, "Missing value for [%s] in config, return default value %s", option_name,
                            missing_value)

            return missing_value

    def _infer_type_from_string(self, text):
        if text == 'True' or text == 'true':
            return True
        elif text == 'False' or text == 'false':
            return False
        return text

    def get_bool_option(self, section, option_name, missing_value=False, subs: Substitutions = None):
        child = section.find(option_name)
        if child is not None:
            value = self._replace_subs(subs, child.text)
            try:
                return self.convert_to_bool(value)
            except Exception:
                pass

        YLogger.warning(self, "Missing value for [%s] in config, return default value %s", option_name,
                        missing_value)
        return missing_value

    def get_int_option(self, section, option_name, missing_value=0, subs=None):
        child = section.find(option_name)
        if child is not None:
            try:
                value = self._replace_subs(subs, child.text)
                return self.convert_to_int(value)

            except Exception as excep:
                pass

        if missing_value is not None:
            YLogger.warning(self, "Missing value for [%s] in config, return default value %d", option_name,
                            missing_value)
        else:
            YLogger.warning(self, "Missing value for [%s] in config, return default value None", option_name)

        return missing_value

    def get_multi_option(self, section, option_name, missing_value=[], subs: Substitutions = None):

        subsections = section.findall(option_name)

        values = []
        for child in subsections:
            if child.text:
                text = child.text
                text = text.strip()
                values.append(self._replace_subs(subs, text))

        if values:
            return values

        return missing_value

    def get_multi_file_option(self, section, option_name, bot_root, missing_value=[], subs: Substitutions = None):

        value = self.get_option(section, option_name, missing_value, subs)

        if isinstance(value, list):
            if value:
                return self._replace_subs(subs, value)

        if isinstance(value, str):
            values = [self._replace_subs(subs, value)]

        else:
            values = []
            if len(value):
                for child in value:
                    if child.tag == "dir":
                        replaced = self._replace_subs(subs, child.text)
                        if replaced:
                            values.append(replaced)

        multis = []
        for value in values:
            multis.append(value.replace('$BOT_ROOT', bot_root))

        return multis



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
from abc import abstractmethod
from programy.utils.logging.ylogger import YLogger
from programy.clients.client import BotClient
from programy.clients.restful.config import RestConfiguration
from programy.clients.restful.apihandlers import APIHandler_V1_0
from programy.clients.restful.apihandlers import APIHandler_V2_0
from programy.clients.restful.apikeys import APIKeysHandler
from programy.clients.restful.auth import RestAuthorizationHandler


class RestBotClient(BotClient, ABC):

    def __init__(self, botid, argument_parser=None):
        BotClient.__init__(self, botid, argument_parser)
        self._api_keys = APIKeysHandler(self.configuration.client_configuration)
        self._authorization = None
        self._v1_0_handler = APIHandler_V1_0(self)
        self._v2_0_handler = APIHandler_V2_0(self)

    @property
    def api_keys(self):
        return self._api_keys

    def get_client_configuration(self):
        return RestConfiguration(self.id)

    def _render_callback(self):
        return False

    def initialise(self):
        self._api_keys.load_api_keys()
        self._authorization = RestAuthorizationHandler.load_authorisation(self)

    def get_variable(self, rest_request, name, method='GET'):
        if method == 'GET':
            if name not in rest_request.args or rest_request.args[name] is None:
                YLogger.error(self, "'%s' missing from GET request", name)
                self.server_abort(message="'%s' missing from GET request"%name, status_code=400)
            if isinstance(rest_request.args[name], list):
                return rest_request.args[name][0]
            else:
                return rest_request.args[name]

        elif method == 'POST':
            if name not in rest_request.json or rest_request.json[name] is None:
                YLogger.error(self, "'%s' missing from POST request", name)
                self.server_abort(message="'%s' missing from POST request"%name, status_code=400)
            return rest_request.json[name]

        else:
            YLogger.error(self, "Invalid REST request type '%s'", method)
            self.server_abort(message="Invalid REST request type '%s'"%method, status_code=400)

    @abstractmethod
    def server_abort(self, message, status_code):
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def create_response(self, response_data, status_code, version=1.0):
        raise NotImplementedError()  # pragma: no cover

    def _get_metadata(self, client_context, metadata):

        if client_context.brain.properties.has_property("fullname"):
            metadata['botName'] = client_context.brain.properties.property("fullname")
        else:
            metadata['botName'] = "Program-y"

        if client_context.brain.properties.has_property("app_version"):
            metadata['version'] = client_context.brain.properties.property("app_version")
        else:
            metadata['version'] = "1.0.0"

        if client_context.brain.properties.has_property("copyright"):
            metadata['copyright'] = client_context.brain.properties.property("copyright")
        else:
            metadata['copyright'] = "Copyright 2016-2020 keithsterling.com"

        if client_context.brain.properties.has_property("botmaster"):
            metadata['authors'] = [client_context.brain.properties.property("botmaster")]
        else:
            metadata['authors'] = ["Keith Sterling"]

    def ask_question(self, userid, question, metadata=None):
        response = ""
        try:
            self._questions += 1
            client_context = self.create_client_context(userid)
            response = client_context.bot.ask_question(client_context, question, responselogger=self)

            if metadata is not None:
                self._get_metadata(client_context, metadata)

        except Exception as e:
            YLogger.exception_nostack(self, "Failed to ask question", e)

        return response

    def process_request(self, request, version=1.0):

        if self._authorization is not None:
            if self._authorization.authorise(request) is False:
                return "Access denied", 403

        if self._api_keys is not None:
            if self._api_keys.use_api_keys():
                if self._api_keys.verify_api_key_usage(request) is False:
                    return 'Unauthorized access', 401

        if version == 1.0:
            return self._v1_0_handler.process_request(self, request)

        elif version == 2.0:
            return self._v2_0_handler.process_request(self, request)

        else:
            return 'Invalid API version', 400

    def dump_request(self, request):
        YLogger.debug(self, str(request))



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.clients.restful.config import RestConfiguration
from programy.utils.substitutions.substitues import Substitutions


class GoogleConfiguration(RestConfiguration):
    DEFAULT_LAUNCH_TEXT = "Hello and welcome"
    DEFAULT_QUIT_TEXT = "Good bye matey"
    DEFAULT_HELP_TEXT = "Ask me anything, I know loads"
    DEFAULT_ERROR_TEXT = "Oopsie there has been an error"

    def __init__(self):
        RestConfiguration.__init__(self, "google")
        self._launch_text = GoogleConfiguration.DEFAULT_LAUNCH_TEXT
        self._launch_srai = None

        self._quit_text = GoogleConfiguration.DEFAULT_QUIT_TEXT
        self._quit_srai = None

        self._help_text = GoogleConfiguration.DEFAULT_HELP_TEXT
        self._help_srai = None

        self._error_text = GoogleConfiguration.DEFAULT_ERROR_TEXT
        self._error_srai = None

    @property
    def launch_text(self):
        return self._launch_text

    @property
    def launch_srai(self):
        return self._launch_srai

    @property
    def quit_text(self):
        return self._quit_text

    @property
    def quit_srai(self):
        return self._quit_srai

    @property
    def help_text(self):
        return self._help_text

    @property
    def help_srai(self):
        return self._help_srai

    @property
    def error_text(self):
        return self._error_text

    @property
    def error_srai(self):
        return self._error_srai

    def load_configuration_section(self, configuration_file, section, bot_root, subs: Substitutions = None):
        assert section is not None

        self._launch_text = configuration_file.get_option(section, "launch_text",
                                                              missing_value=GoogleConfiguration.DEFAULT_LAUNCH_TEXT)
        self._launch_srai = configuration_file.get_option(section, "launch_srai", missing_value=None)

        self._quit_text = configuration_file.get_option(section, "quit_text",
                                                        missing_value=GoogleConfiguration.DEFAULT_QUIT_TEXT)
        self._quit_srai = configuration_file.get_option(section, "quit_srai", missing_value=None)

        self._help_text = configuration_file.get_option(section, "help_text",
                                                        missing_value=GoogleConfiguration.DEFAULT_HELP_TEXT)
        self._help_srai = configuration_file.get_option(section, "help_srai", missing_value=None)

        self._error_text = configuration_file.get_option(section, "error_text",
                                                         missing_value=GoogleConfiguration.DEFAULT_ERROR_TEXT)
        self._error_srai = configuration_file.get_option(section, "error_srai", missing_value=None)

        super(GoogleConfiguration, self).load_configuration_section(configuration_file, section, bot_root,
                                                                    subs=subs)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data["launch_text"] = GoogleConfiguration.DEFAULT_LAUNCH_TEXT
            data["launch_srai"] = None

            data["quit_text"] = GoogleConfiguration.DEFAULT_QUIT_TEXT
            data["quit_srai"] = None

            data["help_text"] = GoogleConfiguration.DEFAULT_HELP_TEXT
            data["help_srai"] = None

            data["error_text"] = GoogleConfiguration.DEFAULT_ERROR_TEXT
            data["error_srai"] = None

        else:
            data["launch_text"] = self._launch_text
            data["launch_srai"] = self._launch_srai

            data["quit_text"] = self._quit_text
            data["quit_srai"] = self._quit_srai

            data["help_text"] = self._help_text
            data["help_srai"] = self._help_srai

            data["error_text"] = self._error_text
            data["error_srai"] = self._error_srai

        super(GoogleConfiguration, self).to_yaml(data, defaults)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from flask import Flask, request
from pymessenger.bot import Bot
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.flask.facebook.config import FacebookConfiguration
from programy.clients.restful.flask.facebook.renderer import FacebookRenderer
from programy.utils.console.console import outputLog


class FacebookBotClient(FlaskRestBotClient):
    
    def __init__(self, argument_parser=None):
        self._access_token = None
        FlaskRestBotClient.__init__(self, 'facebook', argument_parser)

        YLogger.debug(self, "Facebook Client is running....")

        self._facebook_bot = self.create_facebook_bot()

        outputLog(self, "Facebook Client loaded")

    @property
    def facebook_bot(self):
        return self._facebook_bot

    def _render_callback(self):
        return True

    def get_license_keys(self):
        self._access_token = self.license_keys.get_key("FACEBOOK_ACCESS_TOKEN")
        self._verify_token = self.license_keys.get_key("FACEBOOK_VERIFY_TOKEN")

    def get_client_configuration(self):
        return FacebookConfiguration()

    def get_default_renderer(self, callback=True):
        return FacebookRenderer(self)

    def create_facebook_bot(self):
        if self._access_token is not None:
            return Bot(self._access_token)

        YLogger.error(self, "Facebook access token missing, unable to create facebook bot")
        return None

    def get_hub_challenge(self, request):
        return request.args.get("hub.challenge")

    def get_hub_verify_token(self, request):
        return request.args.get("hub.verify_token")

    def verify_fb_token(self, token_sent, request):
        # take token sent by facebook and verify it matches the verify token you sent
        # if they match, allow the request, else return an error
        if token_sent is None:
            YLogger.error(self, "Verify Token is None!")

        if token_sent == self._verify_token:
            return self.get_hub_challenge(request)

        YLogger.error(self, "Facebook verify token failed received [%s]", token_sent)
        return 'Invalid verification token'

    def return_hub_challenge(self, request):
        """Before allowing people to message your bot, Facebook has implemented a verify token
        that confirms all requests that your bot receives came from Facebook."""
        token_sent = self.get_hub_verify_token(request)
        return self.verify_fb_token(token_sent, request)

    def render_response(self, client_context, response):
        # sends user the text message provided via input response parameter
        self._renderer.render(client_context, response)
        return "success"

    def receive_message(self, msg_request):
        if msg_request.method == 'GET':
            return self.return_hub_challenge(msg_request)
        else:
            data = msg_request.get_json()
            self.process_facebook_request(data)
        return "Message Processed"

    def get_recipitent_id(self, message):
        if 'sender' in message:
            if 'id' in message['sender']:
                return message['sender']['id']
        return None

    def get_message_text(self, message):
        if 'message' in message:
            return message['message'].get('text')
        return None

    def get_postback_text(self, message):
        if 'postback' in message:
            return message['postback'].get('payload')
        return None

    def has_attachements(self, message):
        if 'message' in message:
            if message['message'].get('attachments') is not None:
                return True
        return False

    def process_facebook_request(self, request):
        if 'entry' in request:
            for event in request['entry']:
                if 'messaging' in event:
                    messaging = event['messaging']
                    self.process_facebook_message(messaging)

    def process_facebook_message(self, messaging):
        for message in messaging:
            if message.get('message'):
                self.handle_message(message)

            elif message.get('postback'):
                self.handle_postback(message)

    def ask_question(self, client_context, question, metadata=None):
        response = ""
        try:
            self._questions += 1
            response = client_context.bot.ask_question(client_context, question, responselogger=self)

        except Exception as e:
            YLogger.exception(client_context, "Error asking Facebook:", e)

        return response

    def is_echo(self, message):
        if 'message' in message:
            if 'is_echo' in message['message']:
                return message['message']['is_echo']
        return False

    def handle_message(self, message):

        try:
            if self.configuration.client_configuration.debug is True:
                self.dump_request(message)
                self.renderer.print_payload("Incoming", message)

            # Facebook Messenger ID for user so we know where to send response back to
            recipient_id = self.get_recipitent_id(message)
            if recipient_id:

                if self.is_echo(message):
                    YLogger.debug(None, "Handle Facebook is_echo = True")
                    return

                client_context = self.create_client_context(recipient_id)

                message_text = self.get_message_text(message)

                # We have been send a text message, we can respond
                if message_text is not None:
                    response_text = self.handle_text_message(client_context, message_text)

                # else if user sends us a GIF, photo,video, or any other non-text item
                elif self.has_attachements(message):
                    response_text = self.handle_attachement(client_context, message)

                # otherwise its a general error
                else:
                    YLogger.error(client_context, "Facebook general error handling message")
                    response_text = "Sorry, I do not understand you!"

                YLogger.debug(client_context, "Facebook message response: [%s]", response_text)
                self.render_response(client_context, response_text)

        except Exception as e:
            YLogger.exception(None, "Error handling facebook message", e)

    def handle_text_message(self, client_context, message_text):
        YLogger.debug(client_context, "Facebook sent message: [%s]", message_text)
        return self.ask_question(client_context, message_text)

    def handle_attachement(self, client_context, payload):
        try:
            message = payload.get('message')
            attachements = message.get('attachments')
            for attachment in attachements:
                if attachment.get('type') == 'location':
                    return self.handle_location_attachment(client_context, attachment)

        except Exception as e:
            YLogger.exception(client_context, "Unable to handle attachment", e)

        return "Sorry, I cannot handle that type of attachement right now!"

    def handle_location_attachment(self, client_context, attachment):
        try:
            payload = attachment.get('payload')
            coordinates = payload.get('coordinates')
            lat = coordinates.get('lat')
            long = coordinates.get('long')
            question = "RCS XLATLONG LOCATION XLAT %s XLONG %s" % (lat, long)
            return self.ask_question(client_context, question)
        except Exception as e:
            YLogger.exception(client_context, "Unable to handle location attachment", e)

        return "Error processing location!"

    def handle_postback(self, message):

        if self.configuration.client_configuration.debug is True:
            self.dump_request(message)

        # Facebook Messenger ID for user so we know where to send response back to
        recipient_id = self.get_recipitent_id(message)
        if recipient_id:

            client_context = self.create_client_context(recipient_id)

            message_text = self.get_postback_text(message)
            # We have been send a text message, we can respond
            if message_text is not None:
                response_text = self.handle_text_message(client_context, message_text)

            # otherwise its a general error
            else:
                YLogger.error(self, "Facebook general error handling postback")
                response_text = "Sorry, I do not understand you!"

            YLogger.debug(self, "Facebook postback response: [%s]", response_text)
            self.render_response(client_context, response_text)

if __name__ == "__main__":

    outputLog(None, "Initiating Facebook Client...")

    FACEBOOK_CLIENT = FacebookBotClient()

    APP = Flask(__name__)

    @APP.route(FACEBOOK_CLIENT.configuration.client_configuration.api, methods=['GET', 'POST'])
    def receive_message():
        try:
            return FACEBOOK_CLIENT.receive_message(request)
        except Exception as e:
            YLogger.exception(None, "Facebook Error", e)

    FACEBOOK_CLIENT.run(APP)



import csv
import re
import datetime

from programy.clients.client import BotClient
from programy.utils.files.filefinder import FileFinder
from programy.clients.events.console.config import ConsoleConfiguration

class TestQuestion(object):

    def __init__(self, question, answers, topic=None, that=None):
        self._category = None
        self._question = question
        self._answers = answers
        self._answers_regex = []
        self._topic = topic
        self._that = that
        for answer in answers:
            if answer is not None and answer:
                if answer[0] == "!":
                    self._answers_regex.append(("-", re.compile(answer)))
                else:
                    self._answers_regex.append(("+", re.compile(answer)))
        self._response = None

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, category):
        self._category = category

    @property
    def question(self):
        return self._question

    @property
    def answers(self):
        return self._answers

    @property
    def answers_regex(self):
        return self._answers_regex

    @property
    def answers_string(self):
        return " or ".join(self._answers)

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, response):
        self._response = response

    @property
    def topic(self):
        return self._topic

    @property
    def that(self):
        return self._that

class TestFileFileFinder(FileFinder):

    def __init__(self):
        FileFinder.__init__(self)

    def empty_row(self, row):
        return bool(len(row) < 2)

    def is_comment(self, question):
        return bool(question[0] == '#')

    def is_template(self, question):
        return bool(question[0] == '$')

    def clean_up_answer(self, text):
        return text.replace('"', "").strip()

    def add_answers_to_template(self, row, question, templates):
        answers = []
        for answer in row[1:]:
            answers.append(self.clean_up_answer(answer))
        templates[question] = answers

    def add_template_answers(self, templates, answer, answers):
        if answer in templates:
            template = templates[answer]
            for template_answer in template:
                answers.append(template_answer)
        else:
            print("Template [%s] not found!" % answer)

    def load_file_contents(self, filename, userid="*"):
        print("Loading aiml_tests from file [%s]" % filename)
        questions = []
        templates = {}
        with open(filename, 'r') as csvfile:
            csvreader = csv.reader(csvfile, delimiter=',', quotechar='"')
            for row in csvreader:
                if self.empty_row(row) is False:
                    question = row[0]
                    if self.is_comment(question) is False:
                        if self.is_template(question) is True:
                            self.add_answers_to_template(row, question, templates)
                        else:
                            answers = []
                            that = None
                            topic = None
                            for answer in row[1:]:
                                answer = answer.strip()
                                if answer:
                                    if self.is_template(answer) is True:
                                        self.add_template_answers(templates, answer, answers)
                                    else:
                                        if answer.startswith("\"THAT="):
                                            thatsplits = self.clean_up_answer(answer).split("=")
                                            that = thatsplits[1]
                                        elif answer.startswith("\"TOPIC="):
                                            topicsplits = self.clean_up_answer(answer).split("=")
                                            topic = topicsplits[1]
                                        else:
                                            answers.append(self.clean_up_answer(answer))
                            questions.append(TestQuestion(question, answers, topic=topic, that=that))
        return questions


class TestRunnerBotClient(BotClient):

    def __init__(self):
        BotClient.__init__(self, "TestRunner")

    @property
    def test_dir(self):
        return self.arguments.args.test_dir

    @property
    def test_file(self):
        return self.arguments.args.test_file

    @property
    def qna_file(self):
        return self.arguments.args.qna_file

    @property
    def verbose(self):
        return self.arguments.args.verbose

    def get_description(self):
        return 'ProgramY Test Runner Client'

    def add_client_arguments(self, parser=None):
        if parser is not None:
            parser.add_argument('--test_dir', dest='test_dir', help='directory containing test files to run against grammar')
            parser.add_argument('--test_file', dest='test_file', help='Single file of aiml_tests to run against grammar')
            parser.add_argument('--qna_file', dest='qna_file', help='A file containing questions and answers')
            parser.add_argument('--verbose', dest='verbose', action='store_true', help='print out each question to be asked')

    def set_environment(self):
        self.bot.brain.properties.add_property("env", "TestRunner")

    def get_client_configuration(self):
        return ConsoleConfiguration()

    def run(self):
        file_finder = TestFileFileFinder()
        if self.test_dir is not None:
            print("Loading Tests from directory [%s]" % self.test_dir)
            questions = file_finder.load_dir_contents(self.test_dir, extension=".tests", subdir=True)
        else:
            questions = file_finder.load_single_file_contents(self.test_file)

        question_and_answers = open(self.qna_file, "w+")

        successes = []
        failures = []
        warnings = 0
        start = datetime.datetime.now()
        for category in questions.keys():
            for test in questions[category]:
                test.category = category

                if any((c in '$*_^#') for c in test.question):
                    print("WARNING: Wildcards in question! [%s]"%test.question)
                    warnings = warnings +1

                if test.topic is not None:
                    conversation = self.bot.get_conversation(self.clientid)
                    conversation.set_property("topic", test.topic)

                if test.that is not None:
                    response = self.bot.ask_question(self.clientid, test.that, responselogger=self)

                response = self.bot.ask_question(self.clientid, test.question, responselogger=self)
                success = False
                test.response = response

                if self.verbose:
                    print(test.question, "->", test.response)
                question_and_answers.write('"%s", "%s"\n'%(test.question, test.response))

                if not test.answers_regex:
                    if test.response == "":
                        break
                else:
                    for expected_regex in test.answers_regex:
                        regex_type = expected_regex[0]
                        expression = expected_regex[1]
                        match = expression.search(response)
                        if match is not None and regex_type == "+":
                            success = True
                            break
                        elif match is None and regex_type == "-":
                            success = True
                            break

                if success is True:
                    successes.append(test)
                else:
                    failures.append(test)

        question_and_answers.flush ()
        question_and_answers.close ()

        stop = datetime.datetime.now()
        diff = stop-start
        total_tests = len(successes)+len(failures)

        print("Successes: %d" % len(successes))
        print("Failures:  %d" % len(failures))
        if warnings > 0:
            print("Warnings:  %d" % warnings)
        for failure in failures:
            print("\t%s: [%s] expected [%s], got [%s]" % (failure.category, failure.question, failure.answers_string, failure.response))
        print("Total processing time %f.2 secs"%diff.total_seconds())
        print("Thats approx %f aiml_tests per sec"%(total_tests/diff.total_seconds()))

if __name__ == '__main__':

    def run():
        print("Loading, please wait...")
        console_app = TestRunnerBotClient()
        console_app.run()

    run()



import csv
import os
import re
import argparse
import shutil


def tabs(num):
    if tabs == 0:
        return ""
    else:
        return "\t" * num


class VarType(object):
    def __init__(self):
        self._values = []

    def extract_values_between_brackets(self, text, var_type):
        search_str = '%s\((.+?)\)' % var_type
        m = re.search(search_str, text)
        if m:
            return m.group(1)
        else:
            raise Exception("No value for defined type %s", text)


class SelectVar(VarType):
    def __init__(self, text):
        VarType.__init__(self)
        values = self.extract_values_between_brackets(text, "Select")
        self._values = [x.strip() for x in values.split(",")]

    def type_values_to_str(self):
        text = ", ".join(self._values)
        text += " or exit"
        return text

    def get_next_step(self, step, value):
        for condition in step._conditions:
            compare = condition._condition[2:]
            if compare == value:
                return condition._next_step
        return step._conditions[0]._next_step

    def output_template(self, aiml_file, topic_name, step):
        if step._conditions:
            aiml_file.write('\t\t\t\t<condition name="%s">\n' % step._variable)
            for value in self._values:
                next_step = self.get_next_step(step, value)
                aiml_file.write(
                    '\t\t\t\t\t<li value="%s"><srai>%s STEP %s</srai></li>\n' % (value, topic_name, next_step))
            aiml_file.write('\t\t\t\t\t<li><srai>%s STEP %s</srai></li>\n' % (topic_name, step._step))
            aiml_file.write('\t\t\t\t</condition>\n')
        else:
            aiml_file.write('\t\t\t\t<condition name="%s">\n' % step._variable)
            for value in self._values:
                aiml_file.write('\t\t\t\t\t<li value="%s"><srai>EXECUTE %s</srai></li>\n' % (value, topic_name))
            aiml_file.write('\t\t\t\t\t<li><srai>%s STEP %s</srai></li>\n' % (topic_name, step._step))
            aiml_file.write('\t\t\t\t</condition>\n')


class DateVar(VarType):
    def __init__(self, text):
        VarType.__init__(self)
        self._values.append(self.extract_values_between_brackets(text, "Date"))

    def type_values_to_str(self):
        text = self._values[0]
        text += " or exit"
        return text

    def output_template(self, aiml_file, topic_name, step):
        if step._conditions:
            next_step = step._conditions[0]._next_step
            aiml_file.write('\t\t\t\t<think><set name="Valid"><srai>VALID DATE <star /></srai></set></think>\n')
            aiml_file.write('\t\t\t\t<condition name="Valid">\n')
            aiml_file.write('\t\t\t\t\t<li value="TRUE"><srai>%s STEP %s</srai></li>\n' % (topic_name, next_step))
            aiml_file.write('\t\t\t\t\t<li><srai>%s STEP %s</srai></li>\n' % (topic_name, step._step))
            aiml_file.write('\t\t\t\t</condition>\n')
        else:
            aiml_file.write('\t\t\t\t<think><set name="Valid"><srai>VALID DATE <star /></srai></set></think>\n')
            aiml_file.write('\t\t\t\t<condition name="Valid">\n')
            aiml_file.write('\t\t\t\t\t<li value="TRUE"><srai>EXECUTE %s</srai></li>\n' % (topic_name))
            aiml_file.write('\t\t\t\t\t<li><srai>%s STEP %s</srai></li>\n' % (topic_name, step._step))
            aiml_file.write('\t\t\t\t</condition>\n')


class IntVar(VarType):
    def __init__(self, text):
        VarType.__init__(self)
        if '(' in text:
            ranges = self.extract_values_between_brackets(text, "Int")
            splits = ranges.split(",")
            if len(splits) > 1:
                self._values.append(splits[0])  # Min
                self._values.append(splits[1])  # Max
            else:
                self._values.append(splits[0])  # Max

    def type_values_to_str(self):
        if len(self._values) == 2:
            text = " to ".join(self._values)
        else:
            text = " max %s" % self._values[0]
        text += " or exit"
        return text

    def output_template(self, aiml_file, topic_name, step):
        if step._conditions:
            next_step = step._conditions[0]._next_step
            if len(self._values) == 2:
                aiml_file.write(
                    '\t\t\t\t<think><set name="Valid"><srai>VALID INT <star /> %s %s</srai></set></think>\n' % (
                        self._values[0], self._values[1]))
            elif len(self._values) == 1:
                aiml_file.write('\t\t\t\t<think><set name="Valid"><srai>VALID INT <star /> %s</srai></set></think>\n' %
                                self._values[0])
            else:
                aiml_file.write('\t\t\t\t<think><set name="Valid"><srai>VALID INT <star /></srai></set></think>\n')
            aiml_file.write('\t\t\t\t<condition name="Valid">\n')
            aiml_file.write(
                '\t\t\t\t\t<li value="TRUE"><srai>%s STEP %s</srai></li>\n' % (topic_name, next_step))
            aiml_file.write('\t\t\t\t\t<li><srai>%s STEP %s</srai></li>\n' % (topic_name, step._step))
            aiml_file.write('\t\t\t\t</condition>\n')
        else:
            if len(self._values) == 2:
                aiml_file.write(
                    '\t\t\t\t<think><set name="Valid"><srai>VALID INT <star /> %s %s</srai></set></think>\n' % (
                        self._values[0], self._values[1]))
            elif len(self._values) == 1:
                aiml_file.write('\t\t\t\t<think><set name="Valid"><srai>VALID INT <star /> %s</srai></set></think>\n' %
                                self._values[0])
            else:
                aiml_file.write('\t\t\t\t<think><set name="Valid"><srai>VALID INT <star /></srai></set></think>\n')
            aiml_file.write('\t\t\t\t<condition name="Valid">\n')
            aiml_file.write('\t\t\t\t\t<li value="TRUE"><srai>EXECUTE %s</srai></li>\n' % (topic_name))
            aiml_file.write('\t\t\t\t\t<li><srai>%s STEP %s</srai></li>\n' % (topic_name, step._step))
            aiml_file.write('\t\t\t\t</condition>\n')


class TextVar(VarType):
    def __init__(self, text):
        VarType.__init__(self)
        self._values.append(text)

    def type_values_to_str(self):
        return ""

    def output_template(self, aiml_file, topic_name, step):
        if step is not None:
            aiml_file.write('\t\t\t\t<srai>%s STEP %s</srai>\n' % (topic_name, step._next_step))
        else:
            aiml_file.write('\t\t\t\t<srai>EXECUTE %s</srai>\n' % topic_name)


class StepCondition(object):
    def __init__(self, next_step, condition):
        self._next_step = next_step
        self._condition = condition


class Step(object):
    def __init__(self, csv_line):
        print(csv_line)
        self._step = csv_line[0]
        self._prompt = csv_line[1]
        self._variable = csv_line[2]
        self._type = self._var_type(csv_line[3])
        csv_length = len(csv_line)
        self._conditions = []
        if csv_length > 4:
            if len(csv_line[4]) > 0:  # If its more than a blank line
                index = 4
                while index < csv_length:
                    next_step = csv_line[index]
                    condition = csv_line[index + 1]
                    self._conditions.append(StepCondition(next_step, condition))
                    index += 2

    def __str__(self):
        if self._conditions:
            next_steps = ", ".join("%s %s"(x._next_step, x.condition) for x in self._conditions)
        else:
            next_steps = "EXECUTE"
        return "[%s] - %s -> %s" % (self._step, self._prompt, next_steps)

    @staticmethod
    def _var_type(text):
        if text.startswith("Select"):
            return SelectVar(text)
        elif text.startswith("Date"):
            return DateVar(text)
        elif text.startswith("Int"):
            return IntVar(text)
        else:
            return TextVar(text)

    def output_aiml(self, aiml_file, topic_name):

        # Step Question
        aiml_file.write('\t\t<category>\n')
        aiml_file.write('\t\t\t<pattern>\n')
        aiml_file.write('\t\t\t\t%s STEP %s\n' % (topic_name, self._step))
        aiml_file.write('\t\t\t</pattern>\n')
        aiml_file.write('\t\t\t<template>\n')
        if self._type._values:
            aiml_file.write('\t\t\t\t%s - (%s)\n' % (self._prompt, self._type.type_values_to_str()))
        else:
            aiml_file.write('\t\t\t\t%s\n' % self._prompt)
        aiml_file.write('\t\t\t</template>\n')
        aiml_file.write('\t\t</category>\n\n')

        aiml_file.write('\t\t<category>\n')
        aiml_file.write('\t\t\t<pattern>*</pattern>\n')
        if self._type._values:
            aiml_file.write('\t\t\t<that>%s *</that>\n' % self._prompt)
        else:
            aiml_file.write('\t\t\t<that>%s</that>\n' % self._prompt)
        aiml_file.write('\t\t\t<template>\n')
        aiml_file.write('\t\t\t\t<think><set name="%s"><star /></set></think>\n' % self._variable)

        self._type.output_template(aiml_file, topic_name, self)


        aiml_file.write('\t\t\t</template>\n')
        aiml_file.write('\t\t</category>\n\n')

        aiml_file.write('\t\t<category>\n')
        aiml_file.write('\t\t\t<pattern>EXIT</pattern>\n')
        if self._type._values:
            aiml_file.write('\t\t\t<that>%s *</that>\n' % self._prompt)
        else:
            aiml_file.write('\t\t\t<that>%s</that>\n' % self._prompt)
        aiml_file.write('\t\t\t<template>\n')
        aiml_file.write('\t\t\t\t<srai>EXIT %s</srai>\n' % topic_name)
        aiml_file.write('\t\t\t</template>\n')
        aiml_file.write('\t\t</category>\n\n')


class BotFlow(object):
    def __init__(self):
        self._steps = []
        self._name = None

    def load_flow(self, flow_file, name):
        print("Loading flow file [%s]" % flow_file)

        self._steps.clear()
        self._name = name.upper()

        with open(flow_file, "r+") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',', quotechar='"')
            first = True
            for line in csv_reader:
                if first is False:
                    if len(line) > 4:
                        self._steps.append(Step(line))
                    else:
                        print("Invalid line [%s], missing elements" % line)
                first = False

    def write_aiml_header(self, aiml_file):
        aiml_file.write('<?xml version="1" encoding="UTF-8" ?>\n')
        aiml_file.write('<aiml>\n\n')

    def write_aiml_footer(self, aiml_file):
        aiml_file.write('\n</aiml>')

    def write_entry_category(self, aiml_file, first_step):
        aiml_file.write('\t<category>\n')
        aiml_file.write('\t\t<pattern>START %s</pattern>\n' % self._name)
        aiml_file.write('\t\t<template>\n')
        aiml_file.write('\t\t\t<think>\n')
        # Set the topic
        aiml_file.write('\t\t\t\t<set name="topic">%s</set>\n' % self._name)
        # Clear variables before we start
        for step in self._steps:
            aiml_file.write('\t\t\t\t<set name="%s" />\n' % step._variable)
        aiml_file.write('\t\t\t</think>\n')
        # Jump to the first step
        aiml_file.write('\t\t\t<srai>%s STEP %s</srai>\n' % (self._name, first_step))
        aiml_file.write('\t\t</template>\n')
        aiml_file.write('\t</category>\n\n')

    def write_topic_open(self, aiml_file, name):
        aiml_file.write('\t<topic name="%s">\n\n' % name)

    def write_topic_close(self, aiml_file):
        aiml_file.write('\t</topic>\n')

    def write_dialog_flow(self, aiml_file):
        for step in self._steps:
            step.output_aiml(aiml_file, self._name)

    def generate_aiml(self, aiml_dir):
        print("Generating aiml in [%s]" % aiml_dir)

        with open(aiml_dir + os.sep + self._name + ".aiml", "w+", encoding="utf-8") as aiml_file:
            self.write_aiml_header(aiml_file)

            self.write_entry_category(aiml_file, self._steps[0]._step)

            self.write_topic_open(aiml_file, self._name)

            self.write_dialog_flow(aiml_file)

            self.write_topic_close(aiml_file)

            self.write_aiml_footer(aiml_file)

    def copy_supporting_files(self, from_dir, to_dir):
        shutil.copyfile(from_dir + os.sep + "aimlstandardlibrary.aiml", to_dir + os.sep + "aimlstandardlibrary.aiml")
        shutil.copyfile(from_dir + os.sep + "botflowlibrary.aiml", to_dir + os.sep + "botflowlibrary.aiml")


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Program-Y Flow Bot')

    parser.add_argument('-flow', dest='flowfile', help='Flow file to load')
    parser.add_argument('-topic', dest='topic', help='Topic name')
    parser.add_argument('-lib', dest='lib', help='Library of aiml files')
    parser.add_argument('-aiml', dest='aimldir', help='Directory to create aiml files in')

    args = parser.parse_args()

    botflow = BotFlow()

    botflow.load_flow(args.flowfile, args.topic)

    botflow.generate_aiml(args.aimldir)

    botflow.copy_supporting_files(args.lib, args.aimldir)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.pattern.matchcontext import MatchContext


class Sentence:

    def __init__(self, client_context=None, text: str = None,
                 response: str = None,
                 matched_context: MatchContext = None,
                 positivity=0.0, subjectivity=0.5):
        if text is not None:
            self._words = client_context.brain.tokenizer.texts_to_words(text)
        else:
            self._words = []
        self._response = response
        self._matched_context = matched_context
        self._positivity = positivity
        self._subjectivity = subjectivity

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, text: str):
        self._response = text

    @property
    def matched_context(self):
        return self._matched_context

    @matched_context.setter
    def matched_context(self, context):
        self._matched_context = context

    @property
    def words(self):
        return self._words

    @words.setter
    def words(self, words):
        self._words = words[:]

    @property
    def positivity(self):
        return self._positivity

    @positivity.setter
    def positivity(self, value):
        self._positivity = value

    @property
    def subjectivity(self):
        return self._subjectivity

    @subjectivity.setter
    def subjectivity(self, value):
        self._subjectivity = value

    def append_word(self, word):
        self._words.append(word)

    def append_sentence(self, sentence):
        for word in sentence.words:
            self._words.append(word)

    def replace_words(self, client_context, text):
        self._words = Sentence._split_into_words(client_context.brain.tokenizer, text)

    def num_words(self):
        return len(self.words)

    def word(self, num: int):
        if num < self.num_words():
            return self.words[num]
        return None

    def words_from_current_pos(self, client_context, current_pos: int):
        return client_context.brain.tokenizer.words_from_current_pos(self._words, current_pos)

    def text(self, client_context):
        return client_context.brain.tokenizer.words_to_texts(self._words)

    @staticmethod
    def _split_into_words(tokenizer, text):
        if text is None or tokenizer is None:
            return []
        return tokenizer.texts_to_words(text)

    def calculate_sentinment_score(self, client_context):

        assert client_context is not None

        if client_context.bot.sentiment_analyser is not None:
            positivity, subjectivity = client_context.bot.sentiment_analyser.analyse_all(self.text(client_context))
            YLogger.debug(client_context, "Sentiment: positivity[%f], subjectivity [%f]", positivity, subjectivity)
            self._positivity = positivity
            self._subjectivity = subjectivity

    def to_json(self):
        json_data = {
            "words": self._words,
            "response": self._response,
            "positivity": self._positivity,
            "subjectivity": self._subjectivity
        }

        if self._matched_context is not None:
            json_data["matched_context"] = self._matched_context.to_json()

        return json_data

    @staticmethod
    def from_json(client_context, json_data):

        sentence = Sentence(client_context)

        sentence.words = json_data['words']
        sentence.response = json_data['response']
        sentence.positivity = json_data['positivity']
        sentence.subjectivity = json_data['subjectivity']

        if 'matched_context' in json_data:
            sentence.matched_context = MatchContext.from_json(json_data["matched_context"])

        return sentence



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import time

from programy.clients.render.renderer import RichMediaRenderer


class TextRenderer(RichMediaRenderer):

    def __init__(self, callback):
        RichMediaRenderer.__init__(self, callback)

    def _default_output(self):
        return ""

    def _concat_result(self, first, second):
        return first + second

    def handle_text(self, client_context, text):
        if self._client:
            self._client.process_response(client_context, text['text'])
        return text['text']

    def handle_url_button(self, client_context, button):
        rendered = "%s, click %s" % (button['text'], button['url'])
        if self._client:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_postback_button(self, client_context, button):
        if self._client:
            self._client.process_response(client_context, button['postback'])
        return button['postback']

    def handle_link(self, client_context, link):
        rendered = "Open in browser, click %s" % link['url']
        if self._client:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_image(self, client_context, image):
        rendered = "To see the image, click %s" % image['url']
        if self._client:
            self._client.process_response(client_context, rendered)

        return rendered

    def handle_video(self, client_context, video):
        rendered = "To see the video, click %s" % video['url']
        if self._client:
            self._client.process_response(client_context, rendered)

        return rendered

    def _format_card(self, client_context, card):
        del client_context
        rendered = "Image: %s\nTitle: %s\nSubtitle: %s\n" % (card['image'], card['title'], card['subtitle'])
        for button in card['buttons']:
            rendered += "---------------------------------------\n"
            if button['url'] is not None:
                rendered += "%s : %s" % (button['text'], button['url'])
            else:
                rendered += "%s : %s" % (button['text'], button['postback'])
            rendered += "\n---------------------------------------\n"
        return rendered

    def handle_card(self, client_context, card):
        rendered = self._format_card(client_context, card)
        if self._client:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_carousel(self, client_context, carousel):
        rendered = ""
        for card in carousel['cards']:
            rendered += "=========================================\n"
            rendered += self._format_card(client_context, card)
            rendered += "=========================================\n"
        self._client.process_response(client_context, rendered)

    def handle_reply(self, client_context, reply):
        if reply['postback'] is not None:
            if self._client:
                self._client.process_response(client_context, reply['postback'])
                return reply['postback']
        else:
            if self._client:
                self._client.process_response(client_context, reply['text'])
            return reply['text']

    def handle_delay(self, client_context, delay):
        rendered = "..."
        if self._client:
            self._client.process_response(client_context, rendered)
        delay = int(delay['seconds'])
        time.sleep(delay)
        return rendered

    def handle_split(self, client_context, split):
        self._client.process_response(client_context, "\n")
        return "\n"

    def handle_list(self, client_context, lst):
        rendered = ""
        for item in lst.get('items', []):
            rendered += "> %s\n" % item['text']
        self._client.process_response(client_context, rendered)

    def handle_ordered_list(self, client_context, lst):
        rendered = ""
        count = 1
        for item in lst.get('items', []):
            rendered += "%d. %s\n" % (count, item['text'])
            count += 1
        if self._client:
            self._client.process_response(client_context, rendered)
        return rendered

    def handle_location(self, client_context, location):
        del location
        rendered = ""
        if self._client:
            self._client.process_response(client_context, rendered)

        return rendered

    def handle_tts(self, client_context, tts):
        del client_context
        del tts
        return ""



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class Match:
    WORD = 0
    TOPIC = 2
    THAT = 3

    def __init__(self, match_type, node, word):
        self._matched_node_type = match_type

        if node is not None:
            self._matched_node_str = node.to_string(verbose=False)

        self._matched_node_words = []

        if node is not None and \
                (node.is_wildcard() or \
                 node.is_set() or \
                 node.is_iset() or \
                 node.is_bot() or \
                 node.is_regex()):
            self._matched_node_multi_word = True
        else:
            self._matched_node_multi_word = False

        if node is not None:
            self._matched_node_wildcard = node.is_wildcard()
        else:
            self._matched_node_wildcard = False

        if word is not None:
            self.add_word(word)

    def add_word(self, word):
        self._matched_node_words.append(word)

    @property
    def matched_node_type(self):
        return self._matched_node_type

    @matched_node_type.setter
    def matched_node_type(self, node_type):
        self._matched_node_type = node_type

    @property
    def matched_node_str(self):
        return self._matched_node_str

    @matched_node_str.setter
    def matched_node_str(self, node_str):
        self._matched_node_str = node_str

    @property
    def matched_node_multi_word(self):
        return self._matched_node_multi_word

    @matched_node_multi_word.setter
    def matched_node_multi_word(self, multi_word):
        self._matched_node_multi_word = multi_word

    @property
    def matched_node_wildcard(self):
        return self._matched_node_wildcard

    @matched_node_wildcard.setter
    def matched_node_wildcard(self, wildcard):
        self._matched_node_wildcard = wildcard

    @property
    def matched_node_words(self):
        return self._matched_node_words

    @matched_node_words.setter
    def matched_node_words(self, words):
        self._matched_node_words = words[:]

    def joined_words(self, client_context):
        return client_context.brain.tokenizer.words_to_texts(self._matched_node_words)

    @staticmethod
    def type_to_string(match_type):
        if match_type == Match.WORD:
            return "Word"
        elif match_type == Match.TOPIC:
            return "Topic"
        elif match_type == Match.THAT:
            return "That"
        return "Unknown"

    @staticmethod
    def string_to_type(match_type):
        if match_type == 'Word':
            return Match.WORD
        elif match_type == 'Topic':
            return Match.TOPIC
        elif match_type == 'That':
            return Match.THAT
        return -1

    def to_string(self, client_context):
        return "Match=(%s) Node=(%s) Matched=(%s)" % (Match.type_to_string(self.matched_node_type),
                                                      self.matched_node_str,
                                                      self.joined_words(client_context))

    def to_json(self):
        return {"type": Match.type_to_string(self.matched_node_type),
                "node": self.matched_node_str,
                "words": self.matched_node_words,
                "multi_word": self.matched_node_multi_word,
                "wild_card": self.matched_node_wildcard}

    @staticmethod
    def from_json(json_data):
        match = Match(0, None, None)

        match.matched_node_type = Match.string_to_type(json_data["type"])
        match.matched_node_str = json_data["node"]
        match.matched_node_words = json_data["words"]
        match.matched_node_multi_word = json_data["multi_word"]
        match.matched_node_wildcard = json_data["wild_card"]

        return match



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from programy.dialog.tokenizer.tokenizer import Tokenizer


class CjkTokenizer(Tokenizer):

    def __init__(self, split_chars=' '):
        Tokenizer.__init__(self, split_chars)

    @staticmethod
    def _is_chinese_word(word):
        if word is None:
            return False
        else:
            for ch in word:
                if CjkTokenizer._is_chinese_char(ch):
                    return True
            return False

    @staticmethod
    def _is_chinese_char(c):
        r = [
            # 标准CJK文字
            (0x3400, 0x4DB5), (0x4E00, 0x9FA5), (0x9FA6, 0x9FBB), (0xF900, 0xFA2D),
            (0xFA30, 0xFA6A), (0xFA70, 0xFAD9), (0x20000, 0x2A6D6), (0x2F800, 0x2FA1D),
            # 全角ASCII、全角中英文标点、半宽片假名、半宽平假名、半宽韩文字母
            (0xFF00, 0xFFEF),
            # CJK部首补充
            (0x2E80, 0x2EFF),
            # CJK标点符号
            (0x3000, 0x303F),
            # CJK笔划
            (0x31C0, 0x31EF)]
        return any(s <= ord(c) <= e for s, e in r)

    def _is_wildchar(self, ch):
        MATCH_CHARS = ['^', '#', '', '*']
        return bool(ch in MATCH_CHARS)

    def texts_to_words(self, texts):
        if not texts:
            return []

        words = []
        last_word = ''
        for ch in texts:
            if CjkTokenizer._is_chinese_char(ch):
                if len(last_word) > 0:
                    words.append(last_word)
                    last_word = ''
                words.append(ch)
            elif ch == self.split_chars:
                if len(last_word) > 0:
                    words.append(last_word)
                    last_word = ''
            else:
                if self._is_wildchar(ch):
                    if len(last_word) > 0:
                        words.append(last_word)
                        last_word = ''
                    words.append(ch)
                else:
                    last_word += ch

        if len(last_word) > 0:
            words.append(last_word)

        return words

    def words_to_texts(self, words):
        texts = ''

        if words is None:
            words = ['']

        for word in words:
            if CjkTokenizer._is_chinese_word(word):
                texts += word
            elif len(texts) > 0:
                texts += ' ' + word
            else:
                texts += word

        texts = re.sub(r'\s+', ' ', texts)
        stripped_text = texts.strip()
        return stripped_text

    def words_from_current_pos(self, words, current_pos):
        if words:
            return self.words_to_texts(words[current_pos:])
        raise ValueError("Num word array violation !")

    def compare(self, value1, value2):
        cjk_value1 = self.words_to_texts(self.texts_to_words(value1.upper()))
        cjk_value2 = self.words_to_texts(self.texts_to_words(value2.upper()))
        return cjk_value1 == cjk_value2



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from programy.utils.parsing.linenumxml import LineNumberingParser
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-order
from programy.utils.logging.ylogger import YLogger
from programy.storage.entities.store import Store
from programy.parser.exceptions import ParserException
from programy.parser.exceptions import DuplicateGrammarException
from programy.utils.text.text import TextUtils


class CategoryReadOnlyStore(Store):
    WHITESPACE = re.compile('[\n\t\r+]')

    def __init__(self):
        Store.__init__(self)

    def load_all(self, collector):
        raise NotImplementedError("load_all missing from Category Store")  # pragma: no cover

    def load(self, collector, name=None):
        raise NotImplementedError("load missing from Category Store")  # pragma: no cover

    @staticmethod
    def extract_content(name, element):
        content = ET.tostring(element, encoding='utf-8', method='xml').decode('utf-8')
        content = CategoryReadOnlyStore.WHITESPACE.sub('', content)
        content = content.replace('<br>', '<br />')
        content = re.sub(r'\s+', ' ', content)
        start = '<%s>' % name
        end = '</%s>' % name
        firstpos = content.find(start)
        lastpos = content.rfind(end)

        if firstpos == -1 or lastpos == -1:
            return None

        return content[firstpos + len(start):lastpos]

    def find_all(self, element, name, namespace=None):
        if namespace is not None:
            search = "%s:%s"%(list(namespace.keys())[0], name)
            return element.findall(search, namespaces=namespace)
        return element.findall(name)

    def find_element_str(self, name, xml, namespace=None):
        elements = self.find_all(xml, name, namespace)
        if len(elements) > 1:
            raise Exception("Multiple <%s> nodes found in category" % name)

        elif len(elements) == 1:
            return CategoryReadOnlyStore.extract_content(name, elements[0])

        return "*"

    def _load_category(self, groupid, pattern, topic, that, template, parser):

        text = """
<category>
    <pattern>%s</pattern>
    <topic>%s</topic>
    <that>%s</that>
    <template>%s</template>
</category>""" % (pattern, topic, that, template)

        xml = None
        try:
            xml = ET.fromstring(text)
            parser.parse_category(xml, None)

        except DuplicateGrammarException as dupe_excep:
            parser.handle_aiml_duplicate(dupe_excep, groupid, xml)

        except ParserException as parser_excep:
            parser.handle_aiml_error(parser_excep, groupid, xml)

        except Exception as excep:
            YLogger.exception_nostack(self, "Error loading category from db", excep)


class CategoryReadWriteStore(CategoryReadOnlyStore):

    def __init__(self):
        CategoryReadOnlyStore.__init__(self)

    def _parse_topic_expression(self, expression, namespace, groupname, count, success):
        topic = expression.attrib['name']
        for topic_expression in expression:
            that = self.find_element_str("that", topic_expression, namespace)
            pattern = self.find_element_str("pattern", topic_expression, namespace)
            template = self.find_element_str("template", topic_expression, namespace)
            if self.store_category(groupname, "*", topic, that, pattern, template) is True:
                success += 1
            count += 1
        return count, success

    def _parse_category_expression(self, expression, namespace, groupname, count, success):
        topic = self.find_element_str("topic", expression, namespace)
        that = self.find_element_str("that", expression, namespace)
        pattern = self.find_element_str("pattern", expression, namespace)
        template = self.find_element_str("template", expression, namespace)
        if self.store_category(groupname, "*", topic, that, pattern, template) is True:
            success += 1
        count += 1
        return count, success

    def _upload(self, filename):
        count = 0
        success = 0
        groupname = Store.get_just_filename_from_filepath(filename)

        tree = ET.parse(filename, parser=LineNumberingParser())
        aiml = tree.getroot()

        for expression in aiml:
            tag_name, namespace = TextUtils.tag_and_namespace_from_text(expression.tag)

            if tag_name == 'topic':
                count, success = self._parse_topic_expression(expression, namespace, groupname, count, success)

            elif tag_name == 'category':
                count, success = self._parse_category_expression(expression, namespace, groupname, count, success)

            else:
                YLogger.error(self, "Invalid aiml tag type [%s]", tag_name)
                count += 1

        return count, success

    def upload_from_file(self, filename, fileformat=Store.XML_FORMAT, commit=True, verbose=False):
        try:
            return self._upload(filename)

        except Exception as excep:
            YLogger.exception(self, "Failed to load contents of AIML file from [%s]",
                              excep, filename)  # pragma: no cover

        return 0, 0

    def store_category(self, groupid, userid, topic, that, pattern, template):
        raise NotImplementedError("store_category missing from Category Store")  # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.storage.entities.store import Store


class UserStore(Store):

    def __init__(self):
        Store.__init__(self)

    def add_user(self, userid, clientid):
        raise NotImplementedError("add_user missing from User Store")  # pragma: no cover

    def exists(self, userid, clientid):
        raise NotImplementedError("exists missing from User Store")  # pragma: no cover

    def get_links(self, userid):
        raise NotImplementedError("get_links missing from User Store")  # pragma: no cover

    def remove_user(self, userid, clientid):
        raise NotImplementedError("remove_user missing from User Store")  # pragma: no cover

    def remove_user_from_all_clients(self, userid):
        raise NotImplementedError("remove_user_from_all_clients missing from User Store")  # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import threading
import datetime
import requests
from flask import Flask, jsonify, request

from programy.utils.logging.ylogger import YLogger
from programy.clients.ping.config import PingResponderConfig
from programy.utils.console.console import outputLog


class PingResponder():

    def __init__(self, client):
        self._start_time = datetime.datetime.now()
        self._config = client.configuration.client_configuration.responder
        self._client = client

    @property
    def config(self):
        return self._config

    def ping(self):

        payload = {"start_time": "%s" % self._start_time,
                   "client": self._client.id,
                   "questions": self._client.num_questions
                   }
        payload['bots'] = self._client.get_question_counts()
        payload['logging'] = YLogger.snapshot().to_json()

        return payload

    @staticmethod
    def ping_service(ping_app: Flask, config: PingResponderConfig):

        if config.ssl_cert_file is not None and \
                config.ssl_key_file is not None:
            context = (config.ssl_cert_file,
                       config.ssl_key_file)

            outputLog(None, "Healthcheck running in https mode")
            ping_app.run(host=config.host,
                         port=config.port,
                         debug=config.debug,
                         ssl_context=context)
        else:
            outputLog(None, "Healthcheck running in http mode, careful now !")
            ping_app.run(host=config.host,
                         port=config.port,
                         debug=config.debug)

    def start_ping_service(self, ping_app: Flask):
        t = threading.Thread(target=PingResponder.ping_service, args=(ping_app, self.config))
        t.daemon = False
        t.start()
        return t

    def stop_ping_service(self):
        func = request.environ.get('werkzeug.server.shutdown')
        if func is None:
            raise RuntimeError('Not running with the Werkzeug Server')
        func()

    def shutdown_ping_service(self):

        self.unregister_with_healthchecker()

        try:
            url = "http://%s:%d%s" % (self.config.host, self.config.port, self.config.shutdown)
            requests.get(url)

        except Exception:
            YLogger.error(None, "Failed to shutdown ping service")

    def register_with_healthchecker(self):

        if self.config.register is not None:
            if self.config.host is not None:
                host = self.config.host

            else:
                host = self._client.configuration.client_configuration.host

            if self.config.host is not None:
                port = self.config.port

            else:
                port = self._client.configuration.client_configuration.port

            try:
                url = "%s?name=%s&host=%s&port=%s&url=%s" % (self.config.register, self._client.id,
                                                             host, port, self.config.url)
                requests.get(url)

            except Exception as e:
                YLogger.exception(None, "Unable to register with healthchecker", e)

    def unregister_with_healthchecker(self):

        if self.config.unregister is not None:
            try:
                url = "%s?name=%s" % (self.config.unregister, self._client.id)
                requests.get(url)
            except Exception as e:
                outputLog(self, e)
                YLogger.error(None, "Unable to unregister with healthchecker")

    @staticmethod
    def init_ping_response(ping_responder):

        if ping_responder.config.host is None:
            YLogger.info(None, "No REST configuration for ping responder")
            outputLog(None, "Healthcheck now running as part of REST Service...")
            return

        outputLog(None, "Healthcheck now running as separate REST Service...")

        ping_app = Flask(ping_responder.config.name)

        if ping_responder.config.url is not None:
            @ping_app.route(ping_responder.config.url, methods=['GET'])
            def ping():  # pylint: disable=unused-variable
                return jsonify(ping_responder.ping())

        if ping_responder.config.shutdown is not None:
            @ping_app.route(ping_responder.config.shutdown, methods=['GET'])
            def shutdown():  # pylint: disable=unused-variable
                ping_responder.stop_ping_service()
                return 'Server shutting down...'

        ping_responder.start_ping_service(ping_app)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.clients.config import ClientConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class TwitterConfiguration(ClientConfigurationData):

    def __init__(self):
        ClientConfigurationData.__init__(self, "twitter")
        self._description = 'ProgramY AIML2.0 Twitter Client'
        self._polling_interval = 60
        self._rate_limit_sleep = 900

        self._follow_followers = True
        self._respond_to_mentions = True
        self._respond_to_directs = False

        self._mentions = ['#askprogramy']

        self._welcome_message = "Thanks for following me."

    @property
    def polling_interval(self):
        return self._polling_interval

    @property
    def rate_limit_sleep(self):
        return self._rate_limit_sleep

    @property
    def follow_followers(self):
        return self._follow_followers

    @property
    def respond_to_mentions(self):
        return self._respond_to_mentions

    @property
    def respond_to_directs(self):
        return self._respond_to_directs

    @property
    def mentions(self):
        return self._mentions

    @property
    def welcome_message(self):
        return self._welcome_message

    def load_configuration_section(self, configuration_file, section, bot_root, subs: Substitutions = None):
        assert section is not None

        self._polling_interval = configuration_file.get_int_option(section, "polling_interval", missing_value=60,
                                                                   subs=subs)
        self._rate_limit_sleep = configuration_file.get_int_option(section, "rate_limit_sleep", missing_value=900,
                                                                   subs=subs)

        self._follow_followers = configuration_file.get_bool_option(section, "follow_followers", missing_value=False,
                                                                   subs=subs)
        self._respond_to_mentions = configuration_file.get_bool_option(section, "respond_to_mentions", missing_value=False,
                                                                   subs=subs)
        self._respond_to_directs = configuration_file.get_bool_option(section, "respond_to_directs", missing_value=True,
                                                                   subs=subs)

        self._mentions = configuration_file.get_multi_option(section, "mentions", missing_value="", subs=subs)

        self._welcome_message = configuration_file.get_option(section, "welcome_message", subs=subs)

        super(TwitterConfiguration, self).load_configuration_section(configuration_file, section, bot_root,
                                                                     subs=subs)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['polling_interval'] = 60
            data['rate_limit_sleep'] = 900

            data['follow_followers'] = True
            data['respond_to_mentions'] = True
            data['respond_to_directs'] = False

            data['mentions'] = ["#askprogramy"]

            data['welcome_message'] = "Thanks for following me."

            data['storage'] = 'file'
            data['storage_location'] = './storage/twitter.data'

        else:
            data['polling_interval'] = self._polling_interval
            data['rate_limit_sleep'] = self._rate_limit_sleep

            data['follow_followers'] = self._follow_followers
            data['respond_to_mentions'] = self._respond_to_mentions
            data['respond_to_directs'] = self._respond_to_directs

            data['mentions'] = self._mentions[:]

            data['welcome_message'] = self._welcome_message

        super(TwitterConfiguration, self).to_yaml(data, defaults)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import yaml
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.sql.store.sqlstore import SQLStore
from programy.storage.entities.usergroups import UserGroupsStore
from programy.storage.stores.sql.dao.usergroup import AuthoriseUser
from programy.storage.stores.sql.dao.usergroup import UserRole
from programy.storage.stores.sql.dao.usergroup import UserGroup
from programy.storage.stores.sql.dao.usergroup import AuthoriseGroup
from programy.storage.stores.sql.dao.usergroup import GroupUser
from programy.storage.stores.sql.dao.usergroup import GroupGroup
from programy.storage.stores.sql.dao.usergroup import GroupRole
from programy.security.authorise.usergroups import User
from programy.security.authorise.usergroups import Group
from programy.storage.entities.store import Store
from programy.utils.console.console import outputLog


class SQLUserGroupStore(SQLStore, UserGroupsStore):

    def __init__(self, storage_engine):
        SQLStore.__init__(self, storage_engine)
        UserGroupsStore.__init__(self)

    def _get_all(self):
        raise Exception("SQL User groups uses complex multi table data structure, do not call _get_all()")

    def empty(self):
        self._storage_engine.session.query(AuthoriseUser).delete()
        self._storage_engine.session.query(UserRole).delete()
        self._storage_engine.session.query(UserGroup).delete()
        self._storage_engine.session.query(AuthoriseGroup).delete()
        self._storage_engine.session.query(GroupUser).delete()
        self._storage_engine.session.query(GroupRole).delete()

    def load_from_yaml(self, yaml_data, verbose=False):
        self._upload_users(yaml_data, verbose)
        self._upload_groups(yaml_data, verbose)

    def _read_yaml_from_file(self, filename):
        with open(filename, 'r+') as yml_data_file:
            yaml_data = yaml.load(yml_data_file, Loader=yaml.FullLoader)
            self.load_from_yaml(yaml_data)

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):

        try:
            self._read_yaml_from_file(filename)
            self.commit(commit)
            return 1, 1

        except Exception as error:
            YLogger.exception(self, "Failed to load yaml file from [%s]", error, filename)

        return 0, 0

    def _upload_users(self, yaml_data, verbose=False):
        if 'users' in yaml_data:
            for user_name in yaml_data['users'].keys():

                auth_user = AuthoriseUser(name=user_name)
                result = self.storage_engine.session.add(auth_user)

                yaml_obj = yaml_data['users'][user_name]
                if 'roles' in yaml_obj:
                    roles_list = yaml_obj['roles']
                    splits = roles_list.split(",")
                    for role_name in splits:
                        role_name = role_name.strip()
                        user_role = UserRole(user=user_name, role=role_name)
                        if verbose is True:
                            outputLog(self, user_role)
                        self.storage_engine.session.add(user_role)

                if 'groups' in yaml_obj:
                    groups_list = yaml_obj['groups']
                    splits = groups_list.split(",")
                    for group_name in splits:
                        group_name = group_name.strip()
                        user_group = UserGroup(user=user_name, group=group_name)
                        if verbose is True:
                            outputLog(self, user_group)
                        self.storage_engine.session.add(user_group)

    def _upload_groups(self, yaml_data, verbose=False):
        if 'groups' in yaml_data:
            for group_name in yaml_data['groups'].keys():

                auth_group = AuthoriseGroup(name=group_name, parent=None)
                result = self.storage_engine.session.add(auth_group)

                yaml_obj = yaml_data['groups'][group_name]
                if 'roles' in yaml_obj:
                    roles_list = yaml_obj['roles']
                    splits = roles_list.split(",")
                    for role_name in splits:
                        role_name = role_name.strip()
                        group_role = GroupRole(group=group_name, role=role_name)
                        if verbose is True:
                            outputLog(self, group_role)
                        self.storage_engine.session.add(group_role)

                if 'groups' in yaml_obj:
                    groups_list = yaml_obj['groups']
                    splits = groups_list.split(",")
                    for element in splits:
                        inner_group_name = element.strip()
                        group_group = GroupGroup(group=group_name, subgroup=inner_group_name)
                        if verbose is True:
                            outputLog(self, group_group)
                        self.storage_engine.session.add(group_group)

                if 'users' in yaml_obj:
                    users_list = yaml_obj['users']
                    splits = users_list.split(",")
                    for user_name in splits:
                        user_name = user_name.strip()
                        group_user = GroupUser(group=group_name, user=user_name)
                        if verbose is True:
                            outputLog(self, group_user)
                        self.storage_engine.session.add(group_user)

    def load_usergroups(self, usersgroupsauthorisor):

        dbusers = self._storage_engine.session.query(AuthoriseUser)
        for dbuser in dbusers:
            user = User(dbuser.name)
            dbuserroles = self._storage_engine.session.query(UserRole).filter(UserRole.user == dbuser.name)
            for dbuserrole in dbuserroles:
                user.add_role(dbuserrole.role)
            dbusergroups = self._storage_engine.session.query(UserGroup).filter(UserGroup.user == dbuser.name)
            for dbusergroup in dbusergroups:
                user.add_group(dbusergroup.group)
            usersgroupsauthorisor.users[user.userid] = user

        groups = self._storage_engine.session.query(AuthoriseGroup)
        for dbgroup in groups:
            group = Group(dbgroup.name)
            groupusers = self._storage_engine.session.query(GroupUser).filter(GroupUser.group == dbgroup.name)
            for dbgroupuser in groupusers:
                group.add_user(dbgroupuser.user)
            groupgroups = self._storage_engine.session.query(GroupGroup).filter(GroupGroup.group == dbgroup.name)
            for dbgroupgroup in groupgroups:
                group.add_group(dbgroupgroup.subgroup)
            grouproles = self._storage_engine.session.query(GroupRole).filter(GroupRole.group == dbgroup.name)
            for dbgrouprole in grouproles:
                group.add_role(dbgrouprole.role)
            usersgroupsauthorisor.groups[group.groupid] = group

        self._combine_users_and_groups(usersgroupsauthorisor)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.clients.polling.client import PollingBotClient
from programy.clients.polling.xmpp.config import XmppConfiguration
from programy.clients.polling.xmpp.xmpp import XmppClient
from programy.utils.console.console import outputLog


class XmppBotClient(PollingBotClient):

    def __init__(self, argument_parser=None):
        self._xmpp_client = None
        self._server = None
        self._port = None
        self._username = None
        self._password = None
        PollingBotClient.__init__(self, "XMPP", argument_parser)

    def get_client_configuration(self):
        return XmppConfiguration()

    def parse_configuration(self):
        self._server = self.configuration.client_configuration.server
        self._port = self.configuration.client_configuration.port

    def get_license_keys(self):
        self._username = self.license_keys.get_key("XMPP_USERNAME")
        self._password = self.license_keys.get_key("XMPP_PASSWORD")

    def create_client(self, username, password):
        return XmppClient(self, username, password)

    def ask_question(self, userid, question):
        self._questions += 1
        client_context = self.create_client_context(userid)
        response = client_context.bot.ask_question(client_context, question, responselogger=self)
        return response

    def connect(self):
        YLogger.debug(self, "XMPPBotClient Connecting as %s to %s %s", self._username, self._server, self._port)
        self._xmpp_client = self.create_client(self._username, self._password)
        if self._xmpp_client is not None:
            self._xmpp_client.connect((self._server, self._port))
            self._xmpp_client.register_xep_plugins(self.configuration.client_configuration)
            return True
        return False

    def display_connected_message(self):
        outputLog(self, "XMPP Bot connected and running...")

    def poll_and_answer(self):

        running = True
        try:
            print("starting to process")
            self._xmpp_client.process(block=True)
            print("Finished processing")
            outputLog(self, "XMPP client exiting cleanly....")
            running = False

        except KeyboardInterrupt:
            outputLog(self, "XMPP client stopping via keyboard....")
            running = False

        except Exception as excep:
            YLogger.exception(self, "Failed to poll and answer", excep)

        return running


if __name__ == '__main__':

    outputLog(None, "Initiating XMPP Client...")

    xmpp_app = XmppBotClient()
    xmpp_app.run()
